/* FAS Final Async Solution driver for 286/386 versions of System V UNIX */

/* FAS was developed by
Uwe Doering <fas@geminix.in-berlin.de>
Billstedter Pfad 17 b
13591 Berlin
Germany
*/

#if !defined (M_I286)
#ident	"@(#)fas.c	2.12"
#endif

/* Note: This source code has been quite heavily optimized for speed.
         You may wonder that register variables aren't used everywhere.
         This is because there is an overhead in memory accesses
         when using register variables. As you may know data accesses
         usually need much more wait states on the memory bus than
         code accesses (because of page or cache misses). Therefore,
         saving some data accesses has higher priority than saving
         code accesses.

         You may also note some not very elegant constructions that
         may be intentional because they are faster. If you want to
         make style improvements you should check the assembler output
         whether this wouldn't slow things down.

         Decisions for speed optimization were based on assembler
         listings produced by the standard UNIX V [34].X/386 C compiler.

   IMPORTANT: On systems where FAS is to be compiled with the standard
              UNIX C compiler the assembler output file is optimized
              by an AWK script. That is, outb(), inb() and fas_mem_copy()
              calls are substituted with inline i/o mnemonics. As the
              intelligence of this script is limited it is necessary that
              the port address argument of outb() and inb() is a simple
              expression that allows the compiler to produce code that
              fetches the address with a single load operation. Look at
              the assembler output if you are in doubt about an expression.

              This is a kludge but it is the method that produces the
              fastest code with the standard UNIX C compiler.
*/

#include "fas.h"

#if defined (__GNUC__) && !defined (NO_ASM)
#if defined (XENIX)
#define OUTB(port,val) \
({\
	__asm__ volatile ("outb %%al,%%dx" : : "d" ((uint) (port)), "a" ((unchar) (val)));\
})

#define INB(port) \
({\
	unchar __val;\
	__asm__ volatile ("inb %%dx,%%al" : "=a" (__val) : "d" ((uint) (port)));\
	__val;\
})
#else	/* XENIX */
#define OUTB(port,val) \
({\
	__asm__ volatile ("outb (%%dx)" : : "d" ((uint) (port)), "a" ((unchar) (val)));\
})

#define INB(port) \
({\
	unchar __val;\
	__asm__ volatile ("inb (%%dx)" : "=a" (__val) : "d" ((uint) (port)));\
	__val;\
})
#endif	/* XENIX */

#define FAS_MEM_COPY(dst,src,n) \
({\
	__asm__ volatile ("movl %2,%%ecx\n"\
			"\tcld\n"\
			"\tshrl $2,%%ecx\n"\
			"\trep\n"\
			"\tmovsl\n"\
			"\tmovl %2,%%ecx\n"\
			"\tandl $3,%%ecx\n"\
			"\trep\n"\
			"\tmovsb" : : "D" ((unchar *) (dst)),\
						"S" ((unchar *) (src)),\
						"r" ((uint) (n))\
					: "D", "S", "c", "cc");\
})

/* We know better than gcc which variables should be in hard registers.
   This seems to be the only method to force gcc to use a hard register
   for a variable.
*/
#define REG_SI __asm__ ("%esi")
#define REG_DI __asm__ ("%edi")
#define REG_BX __asm__ ("%ebx")
#define REG_CX __asm__ ("%ecx")
#endif	/* __GNUC__ && !NO_ASM */

#if !defined (__GNUC__) && !defined (NO_ASM) && !defined (XENIX)
/* fas_mem_copy () is expanded to inline assembler code by the optimizing
   AWK script (see above).
*/
#define FAS_MEM_COPY(dst,src,n) fas_mem_copy (dst, src, n)
extern void	fas_mem_copy ();

/* Define `OPTIM' for uPort, ISC doesn't know about `.optim', but has turned
   on optimization by default, so we don't need it there anyway.
*/
#if defined (OPTIM)
/* This function inserts the address optimization assembler pseudo-op
   wherever called.
*/

asm	void optim ()
{
	.optim
}

/* This dummy function has nothing to do but to call optim so that
   the `.optim' assembler pseudo-op will be included in the assembler
   file. This must be the first of all functions.
*/

static void
dummy ()
{
	optim ();
}
#endif	/* OPTIM */
#endif	/* !__GNUC__ && !NO_ASM && !XENIX */

#if !defined (REG_SI)
#define REG_SI
#endif
#if !defined (REG_DI)
#define REG_DI
#endif
#if !defined (REG_BX)
#define REG_BX
#endif
#if !defined (REG_CX)
#define REG_CX
#endif

#if defined (NEED_PUT_GETCHAR)
#if defined (SCO) || defined (XENIX)
#define FASPUTCHAR sioputchar
#define FASGETCHAR siogetchar
#else
#define FASPUTCHAR asyputchar
#define FASGETCHAR asygetchar
#endif
#endif

#if defined (PROFILE)
#define STATIC
#else
#define STATIC static
#endif

/* The following defines are used to access multiplexed ports. */
#define FAS_FIRST_CTL(fip,port) \
	((void) (((fip)->device_flags & DF_CTL_FIRST)\
			? OUTB (CTL_PORT.addr, (port).p.ctl)\
			: 0))

#define FAS_EVERY_CTL(fip,port) \
	((void) (((fip)->device_flags & DF_CTL_EVERY)\
			? OUTB (CTL_PORT.addr, (port).p.ctl)\
			: 0))

#define FAS_CTL(fip,port) \
	((void) (((fip)->device_flags & (DF_CTL_FIRST | DF_CTL_EVERY))\
			? OUTB (CTL_PORT.addr, (port).p.ctl)\
			: 0))

#define FAS_FIRST_OUTB(fip,port,val) \
	((void) (((fip)->device_flags & (DF_CTL_FIRST | DF_CTL_EVERY))\
			? OUTB (CTL_PORT.addr, (port).p.ctl)\
			: 0),\
		(void) OUTB ((port).addr, (val)))

#define FAS_OUTB(fip,port,val) \
	((void) (((fip)->device_flags & DF_CTL_EVERY)\
			? OUTB (CTL_PORT.addr, (port).p.ctl)\
			: 0),\
		(void) OUTB ((port).addr, (val)))

#define FAS_SAME_OUTB(fip,port,val) \
	((void) OUTB ((port).addr, (val)))

#define FAS_FIRST_INB(fip,port) \
	((void) (((fip)->device_flags & (DF_CTL_FIRST | DF_CTL_EVERY))\
			? OUTB (CTL_PORT.addr, (port).p.ctl)\
			: 0),\
		INB ((port).addr))

#define FAS_INB(fip,port) \
	((void) (((fip)->device_flags & DF_CTL_EVERY)\
			? OUTB (CTL_PORT.addr, (port).p.ctl)\
			: 0),\
		INB ((port).addr))

#define FAS_SAME_INB(fip,port) \
	(INB ((port).addr))

/* The following defines are used to take apart the minor device numbers. */
#define GET_UNIT(dev)		((dev) & 0x0f)
#define GET_OPEN_MODE(dev)	(fas_open_modes [((dev) >> 4) & 0x0f])

/* Lock device against concurrent use. */
#define GET_DEVICE_LOCK(fip,prio) \
{\
	/* sleep while device is used by an other process */\
	while ((fip)->device_flags & DF_DEVICE_LOCKED)\
		(void) sleep ((caddr_t) &(fip)->tty, (prio));\
	(fip)->device_flags |= DF_DEVICE_LOCKED;\
}

/* Release device. */
#define RELEASE_DEVICE_LOCK(fip) \
{\
	(fip)->device_flags &= ~DF_DEVICE_LOCKED;\
	/* wakeup the process that may wait for this device */\
	(void) wakeup ((caddr_t) &(fip)->tty);\
}

/* Schedule event. */
#define EVENT_SCHED(fip,event) \
{\
	(fip)->event_flags |= (event);\
	if (!event_scheduled)\
	{\
		event_scheduled = TRUE;\
		(void) timeout (fas_event, (caddr_t) NULL, 1);\
	}\
}

/* Put character into receiver ring buffer. */
#define PUT_RECV_BUFFER(fip,ring_ptr,c) \
{\
	*(ring_ptr) = (c);\
	if (++(ring_ptr) == RECV_BUFF_END)\
		(ring_ptr) = RECV_BUFF_BEGIN;\
}

#if !defined (FAS_MEM_COPY)
/* Define the base 2 logarithm of x in the range of 1 to 16 (for shifting). */
#define LOG2(x) \
((x) == 1 ? 0 :\
 (x) == 2 ? 1 :\
 (x) == 4 ? 2 :\
 (x) == 8 ? 3 :\
 (x) == 16 ? 4 : 0)

/* This is the generic version. */
#define FAS_MEM_COPY(dst,src,n) \
{\
	register unchar	*srcptr REG_SI;\
	register unchar	*destptr REG_DI;\
	register uint	num_to_xfer REG_BX;\
\
	num_to_xfer = num_save;\
	srcptr = tmp_srcptr;\
	destptr = tmp_destptr;\
\
	/* do the transfer */\
	if (num_to_xfer >= sizeof (uint) * 2)\
	{\
		/* do multi-byte transfers */\
		num_to_xfer >>= LOG2 (sizeof (uint));\
		do\
		{\
			*((uint *) destptr) = *((uint *) srcptr);\
			srcptr += sizeof (uint);\
			destptr += sizeof (uint);\
		} while (--num_to_xfer);\
\
		num_to_xfer = num_save;\
		if (num_to_xfer &= sizeof (uint) - 1)\
		{\
			/* do the rest with single-byte transfers */\
			do\
			{\
				*destptr = *srcptr;\
				++srcptr;\
				++destptr;\
			} while (--num_to_xfer);\
		}\
	}\
	else\
	{\
		/* do single-byte transfers */\
		do\
		{\
			*destptr = *srcptr;\
			++srcptr;\
			++destptr;\
		} while (--num_to_xfer);\
	}\
}
#endif

/* Calculate highest unsigned value for an integer variable. */
#define UMAX_VAL(var) (((ulong) 1 << (sizeof (var) << 3)) - (ulong) 1)

/* Memory allocation for ring buffers. */
#if !defined (M_I286)
#if defined (XENIX)
#define FAS_MEM_ALLOC(size) sptalloc (size)
#define FAS_MEM_FREE(base,size) sptfree ((caddr_t) (base), (size), 1)
#define FAS_ROUND_UP(size) (size)
#else
#if defined (PG_LOCK)
#define FAS_MEM_ALLOC(size) sptalloc (btoc ((uint) (size)), \
					PG_P | PG_RW | PG_LOCK, NULL, 0)
#else
#define FAS_MEM_ALLOC(size) sptalloc (btoc ((uint) (size)), \
					PG_P | PG_RW, NULL, 0)
#endif
#define FAS_MEM_FREE(base,size) sptfree ((caddr_t) (base), \
						btoc ((uint) (size)), 1)
#define FAS_ROUND_UP(size) ctob (btoc ((uint) (size)))
#endif
#endif

/* Functions provided by this driver. */
void		fasinit P((void));
void		fasopen P((dev_t dev, int flag, int otyp));
void		fasclose P((dev_t dev, int flag, int otyp));
void		fasread P((dev_t dev));
void		faswrite P((dev_t dev));
void		fasproc P((struct tty *ttyp, int cmd));
void		fasioctl P((dev_t dev, int cmd, union ioctl_arg arg, int mode));
void		fasintr P((int vect));
STATIC n_unchar	fas_rproc P((struct fas_internals *fip, n_unchar line_status));
STATIC void	fas_xproc P((struct fas_internals *fip));
STATIC void	fas_open_device P((struct fas_internals *fip));
STATIC void	fas_close_device P((struct fas_internals *fip));
STATIC void	fas_param P((struct fas_internals *fip, int init_type));
STATIC void	fas_pos_by_speed P((struct fas_internals *fipp));
STATIC void	fas_event P((void *dummy));
STATIC void	fas_mproc P((struct fas_internals *fip));
STATIC bool	fas_rxfer P((struct fas_internals *fipp));
STATIC void	fas_xxfer P((struct fas_internals *fipp));
STATIC void	fas_hangup P((struct fas_internals *fip));
STATIC void	fas_timeout P((struct fas_internals *fip));
STATIC void	fas_rdi_enable P((struct fas_internals *fip));
STATIC void	fas_send_xon P((struct fas_internals *fip));
STATIC void	fas_send_xoff P((struct fas_internals *fip));
STATIC uint	fas_make_ctl_val P((struct fas_internals *fip, uint unit,
								uint num));
STATIC void	fas_mem_zero P((unchar *ptr, uint size));
STATIC uint	fas_test_device P((struct fas_internals *fip));
#if defined (NEED_PUT_GETCHAR)
int		FASPUTCHAR P((unchar arg));
int		FASGETCHAR P((void));
#endif
#if defined (NEED_INIT8250)
int		init8250 P((ushort port, ushort ier));
#endif

/* External functions used by this driver. */
extern int	ttinit ();
extern int	ttiocom ();
extern int	ttyflush ();
extern int	SPLINT ();
extern int	SPLWRK ();
extern int	splx ();
extern int	sleep ();
extern int	wakeup ();
extern void	longjmp ();
#if !defined (SVR4)
extern int	signal ();
#endif
extern int	timeout ();
extern int	untimeout ();
#if defined (SCO) || defined (XENIX)
extern int	printcfg ();
#else
extern int	printf ();
#endif
#if !defined (M_I286)
extern caddr_t	sptalloc ();
extern void	sptfree ();
#if !defined (XENIX) && !defined (btoc)
extern uint	btoc ();
#endif
#if !defined (XENIX) && !defined (ctob)
extern uint	ctob ();
#endif
#endif
#if defined (HAVE_VPIX)
extern int	copyin ();
extern int	copyout ();
extern int	v86setint ();
#endif
#if !defined (__GNUC__) || defined (NO_ASM)
#define INB(port) inb (port)
#define OUTB(port,val) outb (port, val)
extern int	inb ();
extern int	outb ();
#endif

/* External data objects used by this driver. */
extern int	tthiwat [];
extern int	ttlowat [];
#if defined (TUNABLE_TTYHOG)
extern int	ttyhog;
#endif

/* The following stuff is defined in `space.c'. */
extern uint	fas_port [];
#if defined (SCO) || defined (XENIX)
extern int	fas_vec [];
#endif
extern uint	fas_modify [];
extern uint	fas_fifo_ctl [];
extern uint	fas_init_seq [];
extern uint	fas_int_ack_seq [];
extern uint	fas_mcb [];
extern ulong	fas_modem [];
extern ulong	fas_flow [];
extern uint	fas_ctl_port [];
extern uint	fas_ctl_val [];
extern uint	fas_pl_select [];
extern uint	fas_layout [] [NUM_UART_REGS];
extern uint	fas_bt_select [];
extern ulong	fas_baud [] [CBAUD + 1];
extern uint	fas_physical_units;
extern uint	fas_port_layouts;
extern uint	fas_baud_tables;
extern struct fas_internals	fas_internals [];
extern struct tty		fas_tty [];
extern struct fas_speed		fas_speed [] [CBAUD + 1];
extern struct fas_internals	*fas_internals_ptr [];
extern struct tty		*fas_tty_ptr [];
extern struct fas_speed		*fas_speed_ptr [];
extern ulong			*fas_baud_ptr [];
/* End of `space.c' references. */

#if defined (NEED_PUT_GETCHAR)
/* Flag to indicate that we have been through fasinit. */
static bool	fas_is_initted = FALSE;
#endif

#if !defined (SCO) && !defined (XENIX)
/* Variables used for the interrupt vector test in the fasinit function. */
static bool	fas_test_mode = FALSE;
static struct fas_internals	*fas_test_fip;
static int	fas_test_vect;
#endif

/* Flag to indicate that the event handler has been scheduled
   via the timeout () function.
*/
static bool	event_scheduled = FALSE;

/* Pointers to the first and last fas_internals structure of the
   interrupt users chain.
*/
static struct fas_internals	*fas_first_int_user
					= (struct fas_internals *) NULL;
static struct fas_internals	*fas_last_int_user
					= (struct fas_internals *) NULL;

/* Threshold for the character transfer to the CLIST buffers. */
static uint	max_rawq_count;

/* Counters for receiver overruns. Each UART type has its own counter
   indexed by the device type.
*/
uint	fas_overrun [NUMBER_OF_TYPES];

/* Counter for temporarily disabled modem status interrupts due
   to noise on the modem status lines.
*/
uint	fas_msi_noise = 0;

/* Lookup table for minor device number -> open mode flags translation. */
static uint	fas_open_modes [16] =
{
	OS_OPEN_FOR_DIALOUT | OS_FAKE_CARRIER_ON | OS_CLOCAL,
	OS_OPEN_FOR_DIALOUT | OS_FAKE_CARRIER_ON | OS_CLOCAL | OS_HWO_HANDSHAKE
							| OS_HWI_HANDSHAKE,
	OS_OPEN_FOR_DIALOUT | OS_FAKE_CARRIER_ON | OS_CLOCAL | OS_HWO_HANDSHAKE,
	OS_OPEN_FOR_DIALOUT | OS_FAKE_CARRIER_ON | OS_CLOCAL | OS_HWO_HANDSHAKE
							| OS_HDX_HANDSHAKE,
	OS_OPEN_FOR_DIALOUT | OS_FAKE_CARRIER_ON,
	OS_OPEN_FOR_DIALOUT | OS_FAKE_CARRIER_ON | OS_HWO_HANDSHAKE
						| OS_HWI_HANDSHAKE,
	OS_OPEN_FOR_DIALOUT | OS_FAKE_CARRIER_ON | OS_HWO_HANDSHAKE,
	OS_OPEN_FOR_DIALOUT | OS_FAKE_CARRIER_ON | OS_HWO_HANDSHAKE
						| OS_HDX_HANDSHAKE,
	OS_OPEN_FOR_DIALIN | OS_WAIT_OPEN | OS_NO_DIALOUT,
	OS_OPEN_FOR_DIALIN | OS_WAIT_OPEN | OS_NO_DIALOUT | OS_HWO_HANDSHAKE
							| OS_HWI_HANDSHAKE,
	OS_OPEN_FOR_DIALIN | OS_WAIT_OPEN | OS_NO_DIALOUT | OS_HWO_HANDSHAKE,
	OS_OPEN_FOR_DIALIN | OS_WAIT_OPEN | OS_NO_DIALOUT | OS_HWO_HANDSHAKE
							| OS_HDX_HANDSHAKE,
	OS_OPEN_FOR_DIALIN | OS_WAIT_OPEN,
	OS_OPEN_FOR_DIALIN | OS_WAIT_OPEN | OS_HWO_HANDSHAKE
						| OS_HWI_HANDSHAKE,
	OS_OPEN_FOR_DIALIN | OS_WAIT_OPEN | OS_HWO_HANDSHAKE,
	OS_OPEN_FOR_DIALIN | OS_WAIT_OPEN | OS_HWO_HANDSHAKE
						| OS_HDX_HANDSHAKE
};

/* Check for the presence of the devices in the fas_port array and if
   the respective device is present, test and initialize it. During the
   initialization the device type is automatically determined and the UART
   is handled according to its requirements.
*/
void
fasinit ()
{
	register struct fas_internals	*fip REG_SI;
	register uint	unit REG_DI;
	uint	logical_units;
	uint	max_rate;
	long	buf_size, min_buf_size;
	struct fas_speed	*speed_ptr;
	ulong	*baud_ptr;
	uint	*offset_ptr;
#if !defined (SCO) && !defined (XENIX)
	int	prev_vect;
#endif
	unchar	port_stat [MAX_UNITS + 1];

#if defined (NEED_PUT_GETCHAR)
	if (fas_is_initted)
		return;

	fas_is_initted = TRUE;
#endif

	{
	    register uint	baud_rate REG_BX;

#if defined (FIX_TTHILOWAT)
	    /* kludge to make sure that the tty buffer high water levels
	       for fast baud rates are high enough to ensure max. throughput
	    */
	    for (max_rate = baud_rate = B0 + 1; baud_rate <= CBAUD; ++baud_rate)
	    {
		if (tthiwat [baud_rate] >= tthiwat [max_rate])
			max_rate = baud_rate;
		else
		{
			tthiwat [baud_rate] = tthiwat [max_rate];
			ttlowat [baud_rate] = ttlowat [max_rate];
		}
	    }
#endif

	    /* calculate baud rate related values */
	    for (unit = 0; unit < fas_baud_tables; ++unit)
	    {
		baud_ptr = fas_baud_ptr [unit] = fas_baud [unit];
		speed_ptr = fas_speed_ptr [unit] = fas_speed [unit];
		speed_ptr [B0].i.max_rate = B0;

		if (!baud_ptr [B0])
			continue;

		for (max_rate = baud_rate = B0 + 1; baud_rate <= CBAUD;
								++baud_rate)
		{
			if (!baud_ptr [baud_rate])
				break;

			speed_ptr [baud_rate].i.ctime
				= ((HZ) * 150L + (baud_ptr [baud_rate] / 2L))
					/ baud_ptr [baud_rate];
			if (speed_ptr [baud_rate].i.ctime < 2)
				speed_ptr [baud_rate].i.ctime = 2;
			speed_ptr [baud_rate].div.val
				= (baud_ptr [B0] + (baud_ptr [baud_rate] / 2L))
					/ baud_ptr [baud_rate];
			if (!speed_ptr [baud_rate].div.val)
				speed_ptr [baud_rate].div.val = 1;
			/* standard xbuf size */
			buf_size = (baud_ptr [baud_rate] + 50L) / 100L;
			/* ideal xbuf size */
			buf_size = (buf_size * (XMIT_BUFF_DRAIN_TIME)
						+ 500L) / 1000L
					- tthiwat [baud_rate];
			/* lower boundary to prevent bottle-neck */
			min_buf_size = (baud_ptr [baud_rate] * (EVENT_TIME)
						+ 25000L) / 50000L;
			if (buf_size < min_buf_size)
				buf_size = min_buf_size;
			if (buf_size < MAX_OUTPUT_FIFO_SIZE * 2L)
				buf_size = MAX_OUTPUT_FIFO_SIZE * 2L;
			if (buf_size > (long) XMIT_BUFF_MAX_SIZE)
				buf_size = XMIT_BUFF_MAX_SIZE;
			if (buf_size > UMAX_VAL (fas_speed [0] [B0].xbuf_size))
				buf_size =
					UMAX_VAL (fas_speed [0] [B0].xbuf_size);
			speed_ptr [baud_rate].xbuf_size = buf_size;
			if (baud_ptr [baud_rate] > baud_ptr [max_rate])
				max_rate = baud_rate;
		}

		/* values for highest baud rate */
		if (baud_rate > CBAUD)
		{
			speed_ptr [B0].div.val = speed_ptr [max_rate].div.val;
			speed_ptr [B0].i.max_rate = max_rate;
		}
	    }
	}

	/* initialize threshold for the character transfer to
	   the CLIST buffers
	*/
#if defined (TUNABLE_TTYHOG)
	max_rawq_count = ttyhog;
#else
	max_rawq_count = TTYHOG;
#endif

	/* execute the init sequence for the serial cards */
	{
	    register uint	*seq_ptr REG_BX;

	    for (seq_ptr = &fas_init_seq [0]; *seq_ptr; seq_ptr += 2)
	    {
		if (*(seq_ptr + 1) & READ_PORT)
			(void) INB (*seq_ptr);
		else
			(void) OUTB (*seq_ptr, *(seq_ptr + 1));
	    }
	}

	/* set overrun counters to zero */
	for (unit = 0; unit < NUMBER_OF_TYPES; ++unit)
		fas_overrun [unit] = 0;

	/* setup the list of pointers to the tty structures */
	for (unit = 0, logical_units = fas_physical_units * 2;
			unit < logical_units; ++unit)
		fas_mem_zero ((unchar *) (fas_tty_ptr [unit] = &fas_tty [unit]),
				sizeof fas_tty [0]);

	/* setup and initialize all serial ports */
	for (unit = 0; unit < fas_physical_units; ++unit)
	{
		fas_internals_ptr [unit] = fip = &fas_internals [unit];
		fas_mem_zero ((unchar *) fip, sizeof fas_internals [0]);
		port_stat [unit] = '-';
		if (fas_port [unit])
		{
			if (fas_bt_select [unit] >= fas_baud_tables)
			{
				port_stat [unit] = '!';
				continue;	/* a configuration error */
			}
			BT_SELECT = fas_bt_select [unit];

			if (fas_speed_ptr [BT_SELECT] [B0].i.max_rate == B0)
			{
				port_stat [unit] = '!';
				continue;	/* a configuration error */
			}

			/* init all of its ports */
			if (fas_ctl_port [unit])
			{
				CTL_PORT.p.addr = fas_ctl_port [unit];

				if (fas_ctl_val [unit] & 0xff00)
					fip->device_flags |= DF_CTL_EVERY;
				else
					fip->device_flags |= DF_CTL_FIRST;
			}

			if (fas_pl_select [unit] >= fas_port_layouts)
			{
				port_stat [unit] = '!';
				continue;	/* a configuration error */
			}

			{
			    register uint	regno REG_BX;

			    offset_ptr = fas_layout [fas_pl_select [unit]];

			    for (regno = 0; regno < NUM_UART_REGS; ++regno)
			    {
				fip->port [regno].p.addr
					= (fip->device_flags & DF_CTL_EVERY)
						? fas_port [unit]
						: fas_port [unit]
							+ offset_ptr [regno];
				fip->port [regno].p.ctl
					= fas_make_ctl_val (fip, unit,
							offset_ptr [regno]);
			    }
			}

			fip->modem.l = fas_modem [unit];
			fip->flow.l = fas_flow [unit];
			fip->po_state = fip->o_state = OS_DEVICE_CLOSED;

			/* mask off invalid bits */
			fip->modem.m.di &= MC_ANY_CONTROL;
			fip->modem.m.eo &= MC_ANY_CONTROL;
			fip->modem.m.ei &= MC_ANY_CONTROL;
			fip->modem.m.ca &= MS_ANY_PRESENT;
			fip->flow.m.ic &= MC_ANY_CONTROL;
			fip->flow.m.oc &= MS_ANY_PRESENT;
			fip->flow.m.oe &= MS_ANY_PRESENT;
			fip->flow.m.hc &= MC_ANY_CONTROL;

			/* set up ring buffer management */
#if !defined (M_I286)
			max_rate = fas_speed_ptr [BT_SELECT] [B0].i.max_rate;
			XMIT_BUFF_SIZE = fas_speed_ptr [BT_SELECT]
							[max_rate].xbuf_size;
			/* standard rbuf size */
			buf_size = (fas_baud_ptr [BT_SELECT] [max_rate]
						+ 50L) / 100L;
			/* ideal rbuf size */
			buf_size = (buf_size * (RECV_BUFF_FILL_TIME)
						+ 500L) / 1000L;
			if (buf_size < (long) RECV_BUFF_MIN_SIZE)
				buf_size = RECV_BUFF_MIN_SIZE;
			if (buf_size > (long) RECV_BUFF_MAX_SIZE)
				buf_size = RECV_BUFF_MAX_SIZE;
			/* if memory is allocated in units > 1 byte, make sure
			   that we use all memory up to the unit boundary
			*/
			RECV_BUFF_SIZE = buf_size
				= FAS_ROUND_UP (buf_size + XMIT_BUFF_SIZE)
					- XMIT_BUFF_SIZE;
#else
			buf_size = RECV_BUFF_MAX_SIZE;
#endif
			RECV_SW_LOW_WATER = buf_size * RECV_PC_SW_LOW_WATER
						/ 100L;
			RECV_SW_HIGH_WATER = buf_size * RECV_PC_SW_HIGH_WATER
						/ 100L;
			RECV_HW_LOW_WATER = buf_size * RECV_PC_HW_LOW_WATER
						/ 100L;
			RECV_HW_HIGH_WATER = buf_size * RECV_PC_HW_HIGH_WATER
						/ 100L;

			/* disable all ints */
			FAS_FIRST_OUTB (fip, INT_ENABLE_PORT, IER = IE_NONE);

			/* is there a serial chip ? */
			if (FAS_SAME_INB (fip, INT_ENABLE_PORT) != IER)
			{
				port_stat [unit] = '?';
				continue;	/* a hardware error */
			}

			/* test the chip thoroughly */
			if (!(fas_modify [unit] & NO_TEST)
				&& (port_stat [unit]
						= (fas_test_device (fip) + '0'))
					!= '0')
				continue;	/* a hardware error */

			FAS_FIRST_OUTB (fip, LINE_CTL_PORT, LCR = 0);
			FAS_OUTB (fip, MDM_CTL_PORT, MCR
					= fas_mcb [unit] | fip->modem.m.di);

			DEVICE_TYPE = TYPE_NS16450;
			fip->xmit_fifo_size = 1;
			port_stat [unit] = '*';

			/* let's see if it's an NS16550A */
			FAS_OUTB (fip, NS_FIFO_CTL_PORT, NS_FIFO_INIT_CMD);
			if (!(~FAS_INB (fip, INT_ID_PORT) & II_NS_FIFO_ENABLED))
			{
				DEVICE_TYPE = TYPE_NS16550A;
				fip->xmit_fifo_size = OUTPUT_NS_FIFO_SIZE;
				port_stat [unit] = 'F';
				FAS_OUTB (fip, NS_FIFO_CTL_PORT, NS_FIFO_CLEAR_CMD);
			}
			else
			{
				FAS_OUTB (fip, NS_FIFO_CTL_PORT, NS_FIFO_CLEAR_CMD);
				/* or is it an i82510 ? */
				FAS_OUTB (fip, I_BANK_PORT, I_BANK_2);
				if (!(~FAS_INB (fip, I_BANK_PORT) & I_BANK_2))
				{
					DEVICE_TYPE = TYPE_I82510;
					fip->xmit_fifo_size = OUTPUT_I_FIFO_SIZE;
					port_stat [unit] = 'f';
					FAS_OUTB (fip, I_BANK_PORT, I_BANK_1);
					FAS_OUTB (fip, I_TCM_PORT, I_FIFO_CLR_XMIT);
					FAS_OUTB (fip, I_RCM_PORT, I_FIFO_CLR_RECV);
				}
				FAS_OUTB (fip, I_BANK_PORT, I_BANK_0);
			}

			/* select FIFO mode */
			if (DEVICE_TYPE != TYPE_NS16450)
			{
				if (fas_fifo_ctl [unit] == FIFO_DEFAULT)
				{
					if (DEVICE_TYPE == TYPE_NS16550A)
					{
#if defined (LOW_INT_LAT)
						FCR = NS_FIFO_SIZE_8
							    | NS_FIFO_ENABLE;
#else
						FCR = NS_FIFO_SIZE_4
							    | NS_FIFO_ENABLE;
#endif
					}
				}
				else if (fas_fifo_ctl [unit]
						== FIFO_EMUL_NS16450)
				{
					fip->xmit_fifo_size = 1;
					if (DEVICE_TYPE == TYPE_NS16550A)
					{
						FCR = NS_FIFO_SIZE_1
							    | NS_FIFO_ENABLE;
					}
				}
				else if (fas_fifo_ctl [unit] == FIFO_OFF
					   || DEVICE_TYPE != TYPE_NS16550A)
				{
					DEVICE_TYPE = TYPE_NS16450;
					fip->xmit_fifo_size = 1;
					port_stat [unit] = '+';
				}
				else switch (fas_fifo_ctl [unit])
				{
					case FIFO_POINTER_DEV:
					case FIFO_TRIGGER_1:
						FCR = NS_FIFO_SIZE_1
							    | NS_FIFO_ENABLE;
						break;
					case FIFO_TRIGGER_4:
						FCR = NS_FIFO_SIZE_4
							    | NS_FIFO_ENABLE;
						break;
					case FIFO_TRIGGER_8:
						FCR = NS_FIFO_SIZE_8
							    | NS_FIFO_ENABLE;
						break;
					case FIFO_TRIGGER_14:
						FCR = NS_FIFO_SIZE_14
							    | NS_FIFO_ENABLE;
						break;
					default:
						DEVICE_TYPE = TYPE_NS16450;
						fip->xmit_fifo_size = 1;
						port_stat [unit] = '+';
						break;
				}
			}

			/* clear potential interrupts */
			(void) FAS_INB (fip, MDM_STATUS_PORT);
			(void) FAS_INB (fip, RCV_DATA_PORT);
			if (FAS_INB (fip, LINE_STATUS_PORT) & LS_RCV_AVAIL)
				(void) FAS_INB (fip, RCV_DATA_PORT);
			(void) FAS_INB (fip, INT_ID_PORT);

			/* do we want the fdx meaning of CTSFLOW/RTSFLOW ? */
			if (fas_modify [unit] & NEW_CTSRTS)
				fip->flow_flags |= FF_NEW_CTSRTS;

			/* do we want hangup protection ? */
			if (!(fas_modify [unit] & NO_HUP_PROTECT))
				fip->device_flags |= DF_HUP_PROTECT;

			/* is the device overrun protected ? */
			if (fas_modify [unit] & NO_OVERRUN)
				fip->device_flags |= DF_NO_OVERRUN;

			/* show that it is present and configured */
			fip->device_flags |= DF_DEVICE_CONFIGURED;
		}
	}

	/* execute the interrupt acknowledge sequence for the serial cards */
	{
	    register uint	*seq_ptr REG_BX;

	    for (seq_ptr = &fas_int_ack_seq [0]; *seq_ptr; seq_ptr += 2)
	    {
		if (*(seq_ptr + 1) & READ_PORT)
			(void) INB (*seq_ptr);
		else
			(void) OUTB (*seq_ptr, *(seq_ptr + 1));
	    }
	}

#if defined (NEED_PUT_GETCHAR)
	fip = &fas_internals [0];
	MCR &= ~fip->modem.m.di;
	FAS_FIRST_OUTB (fip, MDM_CTL_PORT, MCR |= INITIAL_MDM_CONTROL);

	LCR = INITIAL_LINE_CONTROL;
	FAS_OUTB (fip, LINE_CTL_PORT, LCR | LC_ENABLE_DIVISOR);
	FAS_OUTB (fip, DIVISOR_LSB_PORT, fas_speed_ptr [BT_SELECT]
							[INITIAL_BAUD_RATE]
							.div.b.low);
	FAS_OUTB (fip, DIVISOR_MSB_PORT, fas_speed_ptr [BT_SELECT]
							[INITIAL_BAUD_RATE]
							.div.b.high);
	FAS_OUTB (fip, LINE_CTL_PORT, LCR);
#endif

#if !defined (SCO) && !defined (XENIX)
	/* test interrupt vectors */
	for (unit = 0; unit < fas_physical_units; ++unit)
	{
		register uint	rep_count REG_BX;

		fip = fas_internals_ptr [unit];

		if (!(fip->device_flags & DF_DEVICE_CONFIGURED))
			continue;

		prev_vect = -1;
		fas_test_fip = fip;

		/* test the interrupt several times */
		for (rep_count = TEST_INTR_LOOPS; rep_count; --rep_count)
		{
			fas_test_mode = TRUE;	/* select test mode */
			/* Enabling the transmitter holding register empty
			   interrupt generates an immediate interrupt if
			   the transmitter holding register is empty (which
			   is the case at this point). We use this feature
			   to generate the test interrupts.
			*/
			FAS_FIRST_OUTB (fip, INT_ENABLE_PORT,
					IER = IE_XMIT_HOLDING_BUFFER_EMPTY);

			/* make sure that we can get tty interrupts */
			(void) splx (SPLWRK());

			/* disable all ints */
			FAS_FIRST_OUTB (fip, INT_ENABLE_PORT, IER = IE_NONE);

			/* If the interrupt was received and was of the right
			   type, `fas_test_mode' is FALSE now.
			*/
			if (fas_test_mode)
			{
				/* we didn't get an interrupt */
				fas_test_mode = FALSE;
				break;
			}

			/* If we get different vector numbers for the
			   same port there seem to be spurious interrupts.
			*/
			if (prev_vect >= 0 && fas_test_vect != prev_vect)
				break;

			prev_vect = fas_test_vect;
		}

		if (rep_count)
		{
			/* interrupt error */
			fip->device_flags &= ~DF_DEVICE_CONFIGURED;
			port_stat [unit] = '#';
		}
	}
#endif

#if defined (SCO) || defined (XENIX)
	for (unit = 0; unit < fas_physical_units; ++unit)
		(void) printcfg ("fas", fas_port [unit], 7,
					fas_vec [unit], -1,
					"unit=%d type=%c release=2.12.0",
					unit, port_stat [unit]);
#else
	port_stat [unit] = '\0';
	(void) printf ("\nFAS 2.12.0 async driver: Unit 0-%d init state is [%s]\n\n",
			unit - 1,
			port_stat);
#endif
}

/* Open a tty line. This function is called for every open, as opposed
   to the fasclose function which is called only with the last close.
   (called at SPL0)
*/
void
fasopen (dev, flag, otyp)
dev_t	dev;
int	flag;
int	otyp;
{
	register struct fas_internals	*fip REG_SI;
	register struct tty	*ttyp REG_DI;
	register uint	open_mode REG_BX;
	uint	unit;
	bool	have_lock;
	int	old_level;

	/* check for valid port number */
	if ((unit = GET_UNIT (dev)) >= fas_physical_units)
	{
		u.u_error = ENXIO;
		return;
	}

	fip = fas_internals_ptr [unit];

	/* was the port present at init time ? */
	if (!(fip->device_flags & DF_DEVICE_CONFIGURED))
	{
		u.u_error = ENXIO;
		return;
	}

	open_mode = GET_OPEN_MODE (dev);
	have_lock = FALSE;
	old_level = SPLINT ();

	/* loop until we've got the device lock and, owning the lock, we've
	   checked whether the current open mode permits us to open the
	   device
	*/
	for (;;)
	{
		/* If this is a dialin open, but the device is already
		   open for dialout and the FNDELAY flag is not set,
		   wait until the device is closed.
		*/
		if ((open_mode & OS_OPEN_FOR_DIALIN)
				&& (fip->o_state & OS_OPEN_FOR_DIALOUT)
#if defined (FNONBLOCK)
				&& !(flag & (FNDELAY | FNONBLOCK)))
#else
				&& !(flag & FNDELAY))
#endif
		{
			if (have_lock)
			{
				RELEASE_DEVICE_LOCK (fip);
				have_lock = FALSE;
			}
			do
			{
				(void) sleep ((caddr_t) &fip->o_state, TTIPRI);
			} while (fip->o_state & OS_OPEN_FOR_DIALOUT);
		}
		
		/* If the device is already open and another open uses a
		   different open mode or if a dialin process waits for
		   carrier and doesn't allow parallel dialout opens, return
		   with EBUSY error.
		*/
		if ((fip->o_state & ((open_mode & OS_OPEN_FOR_DIALIN)
					? (OS_OPEN_STATES | OS_WAIT_OPEN)
					: (OS_OPEN_STATES | OS_NO_DIALOUT)))
			&& ((flag & FEXCL)
				|| ((open_mode ^ fip->o_state) & (u.u_uid
							? OS_TEST_MASK
							: OS_SU_TEST_MASK))))
		{
			if (have_lock)
				RELEASE_DEVICE_LOCK (fip);
			(void) splx (old_level);
			u.u_error = EBUSY;
			return;
		}

		/* If device is already open and the FAPPEND flag is set,
		   flush the output buffers. This may be used to release
		   processes that got stuck in fasclose () during an
		   exit () call.
		*/
		if ((fip->o_state & OS_OPEN_STATES) && (flag & FAPPEND))
		{
			flag &= ~FAPPEND;	/* flush only once */
			(void) SPLWRK ();
			(void) ttyflush (fip->tty, FWRITE);
			(void) SPLINT ();
		}

		/* if we don't have the device lock, yet, try to get it */
		if (!have_lock)
		{
			if (fip->device_flags & DF_DEVICE_LOCKED)
			{
				GET_DEVICE_LOCK (fip, TTIPRI);
				have_lock = TRUE;
				/* we had to sleep for some time to get the
				   lock, therefore, re-check whether the
				   current open mode still permits us to
				   open the device
				*/
				continue;
			}
			else
				GET_DEVICE_LOCK (fip, TTIPRI);
		}

		break;
	}

	/* disable subsequent opens */
	if (flag & FEXCL)
		open_mode |= OS_EXCLUSIVE_OPEN_1;

	/* set up pointer to tty structure */
	ttyp = (open_mode & OS_OPEN_FOR_DIALIN)
		? fas_tty_ptr [unit + fas_physical_units]
		: fas_tty_ptr [unit];

	/* things to do on first open only */
	if (!(fip->o_state & ((open_mode & OS_OPEN_FOR_DIALIN)
				? (OS_OPEN_STATES | OS_WAIT_OPEN)
				: OS_OPEN_STATES)))
	{
		/* close dialin device before opening dialout device */
		if (fip->o_state & OS_WAIT_OPEN)
		{
			fip->flow_flags &= ~(FF_CD_ENABLED | FF_CARRIER_ON);
			fip->tty->t_state &= ~CARR_ON;
			/* disable receiver data interrupts */
			if (fip->device_flags & DF_RDI_ENABLED)
			{
				FAS_FIRST_OUTB (fip, INT_ENABLE_PORT,
						IER &= ~IE_RECV_DATA_AVAILABLE);
				fip->device_flags &= ~DF_RDI_ENABLED;
			}
			/* re-link fas_internals structure */
			fas_pos_by_speed (fip);
			/* block transmitter */
			fip->device_flags |= DF_XMIT_LOCKED;
			(void) SPLWRK ();
			(void) ttyflush (fip->tty, FREAD | FWRITE);
			(void) SPLINT ();
			fas_close_device (fip);
		}
#if !defined (M_I286)
		else
		{
			/* allocate ring buffer memory */
			(void) SPLWRK ();
			if (!(RECV_BUFF_BEGIN
				= (unchar *) FAS_MEM_ALLOC (RECV_BUFF_SIZE
							     + XMIT_BUFF_SIZE)))
			{
				/* memory allocation failed */
				RELEASE_DEVICE_LOCK (fip);
				(void) splx (old_level);
				u.u_error = ENOMEM;
				return;
			}
			RECV_BUFF_END = XMIT_BUFF_BEGIN = RECV_BUFF_BEGIN
							    + RECV_BUFF_SIZE;
			XMIT_BUFF_END = XMIT_BUFF_BEGIN + XMIT_BUFF_SIZE;
			(void) SPLINT ();
		}
#endif

		/* init data structures */
		fip->tty = ttyp;
		(void) ttinit (ttyp);
		ttyp->t_proc = (int (*)()) fasproc;
		fip->po_state = fip->o_state;
		fip->o_state = open_mode & ~OS_OPEN_STATES;
		fas_open_device (fip);		/* open physical device */

		/* allow pending tty interrupts */
		(void) SPLWRK ();
		(void) SPLINT ();
	}

	/* If dialin open and the FNDELAY flag is not set,
	   block and wait for carrier if device not yet open.
	*/
	if ((open_mode & OS_OPEN_FOR_DIALIN)
#if defined (FNONBLOCK)
			&& !(flag & (FNDELAY | FNONBLOCK)))
#else
			&& !(flag & FNDELAY))
#endif
	{
		/* sleep while open for dialout or no carrier */
		while ((fip->o_state & OS_OPEN_FOR_DIALOUT)
			|| (!(fip->flow_flags & FF_CARRIER_ON)
				&& !(fip->o_state & OS_OPEN_FOR_DIALIN)))
		{
			ttyp->t_state |= WOPEN;
			RELEASE_DEVICE_LOCK (fip);
			(void) sleep ((caddr_t) &ttyp->t_canq, TTIPRI);
			GET_DEVICE_LOCK (fip, TTIPRI);
			ttyp->t_state &= ~WOPEN;
		}
	}

	/* we can't have concurrent ISOPEN and WOPEN states, so
	   wake up waiting dialin processes as we are about to set
	   the ISOPEN flag
	*/
	if (ttyp->t_state & WOPEN)
		(void) wakeup ((caddr_t) &ttyp->t_canq);

	(void) (*linesw [(unchar) ttyp->t_line].l_open) (ttyp);

	/* set open type flags */
	fip->o_state = open_mode;

	RELEASE_DEVICE_LOCK (fip);
	(void) splx (old_level);
}

/* Close a tty line. This is only called if there is no other
   concurrent open left. A blocked dialin open is not counted as
   a concurrent open because in this state it isn't really open.
   (called at SPL0)
*/
void
fasclose (dev, flag, otyp)
dev_t	dev;
int	flag;
int	otyp;
{
	register struct fas_internals	*fip REG_SI;
	register struct tty	*ttyp REG_DI;
	register uint	open_mode REG_BX;
	uint	unit;
	bool	was_signal = FALSE;
	bool	ignore_signal = FALSE;
	int	old_level;

	fip = fas_internals_ptr [unit = GET_UNIT (dev)];

	/* set up pointer to tty structure */
	ttyp = ((open_mode = GET_OPEN_MODE (dev)) & OS_OPEN_FOR_DIALIN)
		? fas_tty_ptr [unit + fas_physical_units]
		: fas_tty_ptr [unit];
	
	old_level = SPLINT ();
	GET_DEVICE_LOCK (fip, PZERO - 1);

	if ((fip->o_state == OS_DEVICE_CLOSED)
			|| ((open_mode & OS_OPEN_FOR_DIALIN)
				? (fip->o_state & OS_OPEN_FOR_DIALOUT)
					&& (fip->po_state == OS_DEVICE_CLOSED)
				: fip->o_state & OS_WAIT_OPEN))
	{
		/* device is not open on the driver level */
		RELEASE_DEVICE_LOCK (fip);
		(void) splx (old_level);
		return;
	}

	if (!((open_mode & OS_OPEN_FOR_DIALIN)
				&& (fip->o_state & OS_OPEN_FOR_DIALOUT)))
	{
		/* wait for buffer drain and catch interrupts */
		while ((fip->flow_flags & FF_CARRIER_ON)
				&& ((fip->flow_flags & FF_OUTPUT_BUSY)
					|| ttyp->t_outq.c_cc
					|| (ttyp->t_state & TIMEOUT)))
		{
			/* if FNDELAY/FNONBLOCK is set, flush output buffers */
#if defined (FNONBLOCK)
			if (flag & (FNDELAY | FNONBLOCK))
			{
				flag &= ~(FNDELAY | FNONBLOCK);
#else
			if (flag & FNDELAY)
			{
				flag &= ~FNDELAY;
#endif
				ignore_signal = TRUE;
				(void) SPLWRK ();
				(void) ttyflush (ttyp, FWRITE);
				(void) SPLINT ();
				continue;
			}
			/* on receipt of a signal, flush output buffers */
			ttyp->t_state |= TTIOW;
			if (sleep ((caddr_t) &ttyp->t_oflag,
						ignore_signal
							? PZERO - 1
							: TTOPRI | PCATCH))
			{
				/* caught signal */
				was_signal = TRUE;
				ignore_signal = TRUE;
				ttyp->t_state &= ~TTIOW;
				(void) SPLWRK ();
				(void) ttyflush (ttyp, FWRITE);
				(void) SPLINT ();
			}
		}

		/* block transmitter and wait until it is
		   empty
		*/
		fip->device_flags |= DF_XMIT_LOCKED;
		while (fip->device_flags & (DF_XMIT_BUSY
						| DF_XMIT_BREAK
						| DF_GUARD_TIMEOUT))
			(void) sleep ((caddr_t) &fip->device_flags,
								PZERO - 1);

		fip->flow_flags &= ~(FF_CD_ENABLED | FF_CARRIER_ON);
		ttyp->t_state &= ~(WOPEN | CARR_ON);
		/* disable receiver data interrupts */
		if (fip->device_flags & DF_RDI_ENABLED)
		{
			FAS_FIRST_OUTB (fip, INT_ENABLE_PORT,
					IER &= ~IE_RECV_DATA_AVAILABLE);
			fip->device_flags &= ~DF_RDI_ENABLED;
		}
		/* re-link fas_internals structure */
		fas_pos_by_speed (fip);
		(void) SPLWRK ();
		(void) ttyflush (ttyp, FREAD | FWRITE);
		(void) SPLINT ();
	}
	else
	{
		fip->flow_flags &= ~(FF_CD_ENABLED | FF_CARRIER_ON);
		ttyp->t_state &= ~(WOPEN | CARR_ON);
	}

	(void) (*linesw [(unchar) ttyp->t_line].l_close) (ttyp);

	if (open_mode & OS_OPEN_FOR_DIALIN)
	{
		if (!(fip->o_state & OS_OPEN_FOR_DIALOUT))
		{
			fas_close_device (fip);
			fip->o_state = OS_DEVICE_CLOSED;
			if (fip->cflag & HUPCL)
			{
				/* request hangup */
				fip->device_flags |= DF_DO_HANGUP;
				(void) timeout (fas_hangup, (caddr_t) fip,
						(HANGUP_DELAY) * (HZ) / 1000);
			}
#if !defined (M_I286)
			/* free ring buffer memory */
			(void) SPLWRK ();
			FAS_MEM_FREE (RECV_BUFF_BEGIN, RECV_BUFF_SIZE
							+ XMIT_BUFF_SIZE);
			(void) SPLINT ();
#endif
		}
		else	/* a parallel dialout open is active */
			fip->po_state = OS_DEVICE_CLOSED;
	}
	else
	{
		fas_close_device (fip);
		fip->o_state = OS_DEVICE_CLOSED;
		if (fip->cflag & HUPCL)
		{
			/* request hangup */
			fip->device_flags |= DF_DO_HANGUP;
			(void) timeout (fas_hangup, (caddr_t) fip,
						(HANGUP_DELAY) * (HZ) / 1000);
		}
		/* If there is a waiting dialin process on
		   this port, reopen the physical device.
		*/
		if (fip->po_state & OS_WAIT_OPEN)
		{
			/* get the dialin version of the
			   tty structure
			*/
			fip->tty = fas_tty_ptr [unit + fas_physical_units];
			fip->o_state = fip->po_state;
			fip->po_state = OS_DEVICE_CLOSED;
			if (!(fip->device_flags & DF_DO_HANGUP))
			{
				/* allow pending tty interrupts */
				(void) SPLWRK ();
				(void) SPLINT ();
				fas_open_device (fip);
			}
		}
#if !defined (M_I286)
		else
		{
			/* free ring buffer memory */
			(void) SPLWRK ();
			FAS_MEM_FREE (RECV_BUFF_BEGIN, RECV_BUFF_SIZE
							+ XMIT_BUFF_SIZE);
			(void) SPLINT ();
		}
#endif

		/* wake up dialin process (if any) */
		(void) wakeup ((caddr_t) &fip->o_state);
	}

	if (!(fip->device_flags & DF_DO_HANGUP))
		RELEASE_DEVICE_LOCK (fip);

	(void) splx (old_level);

	if (was_signal)
#if defined (SVR4)
		longjmp (&u.u_qsav);
#else
		longjmp (u.u_qsav);
#endif
}

/* Read characters from the input buffer.
   (called at SPL0)
*/
void
fasread (dev)
dev_t	dev;
{
	register struct tty	*ttyp REG_CX;

	ttyp = fas_internals_ptr [GET_UNIT (dev)]->tty;

	(void) (*linesw [(unchar) ttyp->t_line].l_read) (ttyp);
}

/* Write characters to the output buffer.
   (called at SPL0)
*/
void
faswrite (dev)
dev_t	dev;
{
	register struct tty	*ttyp REG_CX;

	ttyp = fas_internals_ptr [GET_UNIT (dev)]->tty;

	(void) (*linesw [(unchar) ttyp->t_line].l_write) (ttyp);
}

/* Called by the TTY subsystem (not by FAS itself !) to process various
   device-dependent operations.
   (called at SPL0)
*/
void
fasproc (ttyp, cmd)
register struct tty	*ttyp REG_SI;
int	cmd;
{
	register struct fas_internals	*fip REG_DI;
	uint	unit;
	int	old_level;

	fip = ((unit = ttyp - &fas_tty [0]) >= fas_physical_units)
		? fas_internals_ptr [unit - fas_physical_units]
		: fas_internals_ptr [unit];

	old_level = SPLINT ();

	switch (cmd)
	{
	    case T_TIME:	/* timeout */
		ttyp->t_state &= ~TIMEOUT;
		EVENT_SCHED (fip, EF_DO_XXFER);
		break;

	    case T_OUTPUT:	/* output characters to the transmitter */
		if (fip->xmit_ring_max > fip->xmit_ring_cnt)
			EVENT_SCHED (fip, EF_DO_XXFER);
		break;

	    case T_SUSPEND:	/* suspend character output */
		fip->flow_flags |= FF_SWO_STOPPED;
		ttyp->t_state |= TTSTOP;
		break;

	    case T_RESUME:	/* restart character output */
		fip->flow_flags &= ~FF_SWO_STOPPED;
		ttyp->t_state &= ~TTSTOP;
		fas_xproc (fip);
		break;

	    case T_BLOCK:	/* stop character input, request XOFF */
		ttyp->t_state |= TBLOCK;
		break;	/* note: we do our own XON/XOFF */

	    case T_UNBLOCK:	/* restart character input, request XON */
		ttyp->t_state &= ~TBLOCK;
		break;	/* note: we do our own XON/XOFF */

	    case T_RFLUSH:	/* flush input buffers and restart input */
		if (DEVICE_TYPE == TYPE_NS16550A)
			FAS_FIRST_OUTB (fip, NS_FIFO_CTL_PORT, FCR
						| NS_FIFO_CLR_RECV);
		else if (DEVICE_TYPE == TYPE_I82510)
		{
			FAS_FIRST_OUTB (fip, I_BANK_PORT, I_BANK_1);
			FAS_OUTB (fip, I_RCM_PORT, I_FIFO_CLR_RECV);
			FAS_OUTB (fip, I_BANK_PORT, I_BANK_0);
		}
		(void) FAS_FIRST_INB (fip, RCV_DATA_PORT);
		if (FAS_INB (fip, LINE_STATUS_PORT) & LS_RCV_AVAIL)
			(void) FAS_INB (fip, RCV_DATA_PORT);

		fip->recv_ring_put_ptr = fip->recv_ring_take_ptr
					= RECV_BUFF_BEGIN;
		fip->recv_ring_cnt = 0;
		ttyp->t_state &= ~TBLOCK;

		/* restart input */
		if ((fip->flow_flags & (FF_HWI_HANDSHAKE | FF_HWI_STARTED))
				== FF_HWI_HANDSHAKE)
		{
			FAS_OUTB (fip, MDM_CTL_PORT,
						MCR |= fip->flow.m.ic);
			fip->flow_flags |= FF_HWI_STARTED;
		}
		if ((fip->iflag & IXOFF)
				&& (fip->flow_flags & FF_SWI_STOPPED))
			fas_send_xon (fip);
		break;

	    case T_WFLUSH:	/* flush output buffer and restart output */
		if (DEVICE_TYPE == TYPE_NS16550A)
			FAS_FIRST_OUTB (fip, NS_FIFO_CTL_PORT, FCR
						| NS_FIFO_CLR_XMIT);
		else if (DEVICE_TYPE == TYPE_I82510)
		{
			FAS_FIRST_OUTB (fip, I_BANK_PORT, I_BANK_1);
			FAS_OUTB (fip, I_TCM_PORT, I_FIFO_CLR_XMIT);
			FAS_OUTB (fip, I_BANK_PORT, I_BANK_0);
		}

		fip->xmit_ring_put_ptr = fip->xmit_ring_take_ptr
					= XMIT_BUFF_BEGIN;
		fip->xmit_ring_cnt = 0;

		ttyp->t_tbuf.c_size -= ttyp->t_tbuf.c_count;
		ttyp->t_tbuf.c_count = 0;

		fip->flow_flags &= ~(FF_SWO_STOPPED | FF_SW_FC_REQ);
		ttyp->t_state &= ~(TTSTOP | TTXON | TTXOFF);

		if ((fip->flow_flags & FF_OUTPUT_BUSY)
			&& !(fip->device_flags & (DF_XMIT_BUSY
							| DF_GUARD_TIMEOUT
							| DF_XMIT_BREAK)))
		{
			fip->flow_flags &= ~FF_OUTPUT_BUSY;
			ttyp->t_state &= ~BUSY;
			/* check half duplex output flow control */
			if ((fip->flow_flags & (FF_HDX_HANDSHAKE
							| FF_HDX_STARTED))
				== (FF_HDX_HANDSHAKE | FF_HDX_STARTED))
			{
				FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR &= ~fip->flow.m.hc);
				fip->flow_flags &= ~FF_HDX_STARTED;
			}
			EVENT_SCHED (fip, EF_DO_XXFER);
		}
		break;

	    case T_BREAK:	/* do a break on the transmitter line */
		fip->device_flags |= DF_XMIT_BREAK;
		fip->flow_flags |= FF_OUTPUT_BUSY;
		ttyp->t_state |= BUSY;
		/* check half duplex output flow control */
		if ((fip->flow_flags & (FF_HDX_HANDSHAKE | FF_HDX_STARTED))
			== FF_HDX_HANDSHAKE)
		{
			FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR |= fip->flow.m.hc);
			fip->flow_flags |= FF_HDX_STARTED;
		}
		if (fip->device_flags & (DF_XMIT_BUSY | DF_GUARD_TIMEOUT))
		{
			fip->device_flags |= DF_DO_BREAK;
		}
		else
		{
			/* set up break request flags */
			FAS_FIRST_OUTB (fip, LINE_CTL_PORT,
						LCR |= LC_SET_BREAK_LEVEL);
			(void) timeout (fas_timeout, (caddr_t) fip,
						(BREAK_TIME) * (HZ) / 1000);
		}
		break;

	    case T_PARM:	/* set up port according to termio structure */
		fas_param (fip, SOFT_INIT);
		break;
	}

	(void) splx (old_level);
}

/* Process ioctl calls.
   (called at SPL0)
*/
void
fasioctl (dev, cmd, arg, mode)
dev_t	dev;
int	cmd;
union ioctl_arg	arg;
int	mode;
{
	register struct fas_internals	*fip REG_SI;
	register struct tty	*ttyp REG_DI;
#if defined (HAVE_VPIX)
	register int	v86_cmd REG_BX;
	int	v86_data;
#if defined (SCO) || defined (XENIX)
	struct termss	termss;
#endif
	n_unchar	vpix_status;
	unchar	cpbyte;
#endif
	int	old_level;

	fip = fas_internals_ptr [GET_UNIT (dev)];
	ttyp = fip->tty;

	/* process ioctl commands */
	switch (cmd)
	{
#if defined (HAVE_VPIX)
		case AIOCINTTYPE:	/* set pseudorupt type */
			switch (arg.iarg)
			{
				case V86VI_KBD:
				case V86VI_SERIAL0:
				case V86VI_SERIAL1:
					old_level = SPLINT ();
					fip->v86_intmask = arg.iarg;
					(void) splx (old_level);
					break;

				default:
					old_level = SPLINT ();
					fip->v86_intmask = V86VI_SERIAL0;
					(void) splx (old_level);
					break;
			}
			break;

		case AIOCDOSMODE:	/* enable dos mode */
			if (!(fip->iflag & DOSMODE))
			{
				old_level = SPLINT ();
				fip->v86_proc = u.u_procp->p_v86;
				if (!(fip->v86_intmask))
					fip->v86_intmask = V86VI_SERIAL0;
				ttyp->t_iflag |= DOSMODE;
				if (fip->v86_intmask != V86VI_KBD)
					ttyp->t_cflag |= CLOCAL;
				fas_param (fip, SOFT_INIT);
				(void) splx (old_level);
			}
			u.u_rval1 = 0;
			break;

		case AIOCNONDOSMODE:	/* disable dos mode */
			if (fip->iflag & DOSMODE)
			{
				old_level = SPLINT ();
				fip->v86_proc = (v86_t *) NULL;
				fip->v86_intmask = 0;
				ttyp->t_iflag &= ~DOSMODE;
				if (fip->flow_flags & FF_RXFER_STOPPED)
				{
					fip->flow_flags &= ~FF_RXFER_STOPPED;
					/* schedule character transfer
					   to CLIST buffer
					*/
					if (fip->recv_ring_cnt)
						EVENT_SCHED (fip, EF_DO_RXFER);
				}
				LCR = 0;
				fas_param (fip, HARD_INIT);
				(void) splx (old_level);
			}
			u.u_rval1 = 0;
			break;

		case AIOCSERIALOUT:	/* setup port registers for dos */
			if (fip->iflag & DOSMODE)
			{
				/* wait until output is done */
				old_level = SPLINT ();
				while ((fip->flow_flags & FF_OUTPUT_BUSY)
					|| ttyp->t_outq.c_cc
					|| (ttyp->t_state & TIMEOUT))
				{
					ttyp->t_state |= TTIOW;
					(void) sleep ((caddr_t) &ttyp->t_oflag,
									TTOPRI);
				}

				/* block transmitter and wait until it is
				   empty
				*/
				fip->device_flags |= DF_XMIT_LOCKED;
				while (fip->device_flags & (DF_XMIT_BUSY
							| DF_XMIT_BREAK
							| DF_GUARD_TIMEOUT))
					(void) sleep ((caddr_t) &fip->
								device_flags,
							PZERO - 1);
				(void) splx (old_level);

				/* get port write command */
				if (copyin (arg.cparg, &cpbyte, sizeof cpbyte)
						== -1)
				{
					u.u_error = EFAULT;
					goto restart_output;
				}
				v86_cmd = cpbyte;
				/* set divisor lsb requested */
				if (v86_cmd & SIO_MASK (SO_DIVLLSB))
				{
					if (copyin (arg.cparg + SO_DIVLLSB,
							&cpbyte, sizeof cpbyte)
						== -1)
					{
						u.u_error = EFAULT;
						goto restart_output;
					}
					v86_data = cpbyte;
					old_level = SPLINT ();
					FAS_FIRST_OUTB (fip, LINE_CTL_PORT, LCR
							| LC_ENABLE_DIVISOR);
					FAS_OUTB (fip, DIVISOR_LSB_PORT, v86_data);
					FAS_OUTB (fip, LINE_CTL_PORT, LCR);
					(void) splx (old_level);
				}
				/* set divisor msb requested */
				if (v86_cmd & SIO_MASK (SO_DIVLMSB))
				{
					if (copyin (arg.cparg + SO_DIVLMSB,
							&cpbyte, sizeof cpbyte)
						== -1)
					{
						u.u_error = EFAULT;
						goto restart_output;
					}
					v86_data = cpbyte;
					old_level = SPLINT ();
					FAS_FIRST_OUTB (fip, LINE_CTL_PORT, LCR
							| LC_ENABLE_DIVISOR);
					FAS_OUTB (fip, DIVISOR_MSB_PORT, v86_data);
					FAS_OUTB (fip, LINE_CTL_PORT, LCR);
					(void) splx (old_level);
				}
				/* set lcr requested */
				if (v86_cmd & SIO_MASK (SO_LCR))
				{
					if (copyin (arg.cparg + SO_LCR,
							&cpbyte, sizeof cpbyte)
						== -1)
					{
						u.u_error = EFAULT;
						goto restart_output;
					}
					v86_data = cpbyte;
					old_level = SPLINT ();
					FAS_FIRST_OUTB (fip, LINE_CTL_PORT,
							LCR = v86_data
							  & ~LC_ENABLE_DIVISOR);
					(void) splx (old_level);
				}
				/* set mcr requested */
				if (v86_cmd & SIO_MASK (SO_MCR))
				{
					if (copyin (arg.cparg + SO_MCR,
							&cpbyte, sizeof cpbyte)
						== -1)
					{
						u.u_error = EFAULT;
						goto restart_output;
					}
					v86_data = cpbyte;
					old_level = SPLINT ();
					/* virtual dtr processing */
					if ((v86_data & MC_SET_DTR)
						&& (fip->flow_flags
							& FF_CD_ENABLED))
					{
						MCR |= (fip->o_state
								& OS_WAIT_OPEN)
							? fip->modem.m.ei
							: fip->modem.m.eo;
					}
					else
					{
						MCR &= (fip->o_state
								& OS_WAIT_OPEN)
							? ~fip->modem.m.ei
							: ~fip->modem.m.eo;
					}
					/* virtual rts processing */
					if (fip->flow_flags
							& FF_HWI_HANDSHAKE)
					{
					  if (v86_data & MC_SET_RTS)
					  {
						if (fip->flow_flags
							& FF_RXFER_STOPPED)
						{
						  fip->flow_flags
							&= ~FF_RXFER_STOPPED;
						  /* schedule character transfer
						     to CLIST buffer
						  */
						  if (fip->recv_ring_cnt)
							EVENT_SCHED (fip,
								EF_DO_RXFER);
						}
					  }
					  else
						fip->flow_flags
							|= FF_RXFER_STOPPED;
					}
					else if (!(fip->flow_flags
							& FF_HDX_HANDSHAKE))
					{
						if (v86_data & MC_SET_RTS)
							MCR |= fip->flow.m.hc;
						else
							MCR &= ~fip->flow.m.hc;
					}
					FAS_FIRST_OUTB (fip, MDM_CTL_PORT, MCR);
					(void) splx (old_level);
				}

restart_output:
				old_level = SPLINT ();
				/* enable transmitter and restart output */
				fip->device_flags &= ~DF_XMIT_LOCKED;
				fas_xproc (fip);
				(void) splx (old_level);
			}
			break;

		case AIOCSERIALIN:	/* read port registers for dos */
			if (fip->iflag & DOSMODE)
			{
				if (copyin (arg.cparg, &cpbyte, sizeof cpbyte)
					== -1)
				{
					u.u_error = EFAULT;
					break;
				}
				v86_cmd = cpbyte;
				if (v86_cmd & SIO_MASK (SI_MSR))
				{
					vpix_status = MSR & MS_ANY_PRESENT;
					if (fip->flow_flags
						& FF_HWO_HANDSHAKE)
					    vpix_status |= fip->flow.m.oc
							    | fip->flow.m.oe;
					if (!(fip->cflag & CLOCAL))
					    vpix_status |= fip->modem.m.ca;
					cpbyte = vpix_status;
					if (copyout (&cpbyte,
							arg.cparg + SI_MSR,
							sizeof cpbyte)
						== -1)
					{
						u.u_error = EFAULT;
						break;
					}
				}
			}
			break;

		case AIOCSETSS:	/* set start/stop characters */
#if defined (SCO) || defined (XENIX)
			/* SCO UNIX and Xenix pass a pointer to struct termss */
			if (copyin (arg.cparg, &termss, sizeof termss)
					== -1)
			{
				u.u_error = EFAULT;
				break;
			}
#define GET_TERMSS(member)	(termss.member)
#else
			/* The rest of the world passes struct termss */
#define GET_TERMSS(member)	(((struct termss *) &arg.iarg)->member)
#endif
			old_level = SPLINT ();
			fip->start_char = GET_TERMSS (ss_start);
			fip->stop_char = GET_TERMSS (ss_stop);
			(void) splx (old_level);
			break;

		case AIOCINFO:	/* show what type of device we are */
			u.u_rval1 = ('a' << 8) | GET_UNIT (dev);
			break;
#endif
#if defined (RTS_TOG)
		case RTS_TOG:	/* set hardware handshake output */
			if (arg.iarg == 0)
			{
			    /* low level */
			    old_level = SPLINT ();
			    /* set hardware handshake output only if unused
			       otherwise
			    */
			    if (!(fip->flow_flags & (FF_HWI_HANDSHAKE
							| FF_HDX_HANDSHAKE
							| FF_DEF_HHO_LOW))
#if defined (HAVE_VPIX)
					&& !(fip->iflag & DOSMODE)
#endif
					)
			    {
				/* block transmitter and wait until it is
				   empty
				*/
				fip->device_flags |= DF_XMIT_LOCKED;
				while (fip->device_flags & (DF_XMIT_BUSY
							| DF_XMIT_BREAK
							| DF_GUARD_TIMEOUT))
					(void) sleep ((caddr_t) &fip->
								device_flags,
							PZERO - 1);
	    			/* suspend character output */
				fip->flow_flags |= FF_SWO_STOPPED;
				ttyp->t_state |= TTSTOP;
				FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR &= ~fip->flow.m.hc);
				/* enable transmitter and restart output */
				fip->device_flags &= ~DF_XMIT_LOCKED;
				fas_xproc (fip);
			    }
			    fip->flow_flags |= FF_DEF_HHO_LOW;
			    (void) splx (old_level);
			}
			else if (arg.iarg == 1)
			{
			    /* high level */
			    old_level = SPLINT ();
			    /* set hardware handshake output only if unused
			       otherwise
			    */
			    if ((fip->flow_flags & (FF_HWI_HANDSHAKE
							| FF_HDX_HANDSHAKE
							| FF_DEF_HHO_LOW))
					== FF_DEF_HHO_LOW
#if defined (HAVE_VPIX)
					&& !(fip->iflag & DOSMODE)
#endif
					)
			    {
	    			/* restart character output */
				fip->flow_flags &= ~FF_SWO_STOPPED;
				ttyp->t_state &= ~TTSTOP;
				FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR |= fip->flow.m.hc);
				fas_xproc (fip);
			    }
			    fip->flow_flags &= ~FF_DEF_HHO_LOW;
			    (void) splx (old_level);
			}
			break;
#endif
		default:	/* default ioctl processing */
			/* If it is a TCSETA* command, call fas_param ().
			   There is a bug in ttiocom with TCSELE. It
			   always tells the tty driver to reprogram the
			   port. This will cause lost input characters.
			   So we don't do it. SCO UNIX and Xenix use
			   IOC_SELECT instead of TCSELE!
			*/
			old_level = SPLWRK ();
#if defined (TCSELE)
			if (ttiocom (ttyp, cmd, arg, mode) && cmd != TCSELE)
#else
#if defined (IOC_SELECT)
			if (ttiocom (ttyp, cmd, arg, mode)
						&& cmd != IOC_SELECT)
#else
			if (ttiocom (ttyp, cmd, arg, mode))
#endif
#endif
			{
				/* reprogram the port */
				(void) SPLINT ();
				fas_param (fip, SOFT_INIT);
			}
			(void) splx (old_level);
			break;
	}
}

/* Main FAS interrupt handler. Actual character processing is splitted
   into sub-functions.
   (called at SPLINT)
*/
void
fasintr (vect)
int	vect;
{
	register struct fas_internals	*fip REG_SI;
	uint	num_polls;
	bool	was_rint, was_xint, was_mint, do_xproc;

	fip = fas_first_int_user;

	if (fip)
	{
	    register uint	poll_cnt REG_DI;

	    was_rint = FALSE;
	    was_xint = FALSE;
	    was_mint = FALSE;

	    poll_cnt = 1;

	    /* Loop through all users of the interrupt users chain
	       and look for characters in the receiver buffer.
	       We poll the line status register regardless of
	       whether the UART has interrupted or not. This
	       synchronizes concurrently running receiver FIFOs
	       and dramatically reduces the overall interrupt
	       frequency.
	    */
	    do
	    {
		register n_unchar	lstatus REG_BX;
fastloop1:
		/* read in all characters from all UARTs */
		if (!(fip->device_flags & DF_RDI_ENABLED)
			|| !((lstatus = FAS_FIRST_INB (fip, LINE_STATUS_PORT))
					& LS_RCV_AVAIL))
		{
			/* speed beats beauty */
			fip = fip->next_int_user;
			++poll_cnt;
			if (fip)
				goto fastloop1;
			break;
		}

		/* do the character processing */
		(void) fas_rproc (fip, lstatus);
		was_rint = TRUE;
		fip = fip->next_int_user;
		++poll_cnt;
	    } while (fip);

	    /* Save the number of UARTs we have to poll to be sure that
	       there are no further pending interrupts.
	    */
	    num_polls = poll_cnt;

	    /* The 8259 interrupt controller is set up for edge trigger.
	       Therefore, we must loop until we make a complete pass without
	       getting any UARTs that are interrupting.
	    */
	    for (;;)
	    {
fastloop2:
		if (!fip)
			fip = fas_first_int_user;	/* wrap around */

		/* process only ports that have an interrupt
		   pending
		*/
		if (FAS_FIRST_INB (fip, INT_ID_PORT) & II_NO_INTS_PENDING)
		{
			/* speed beats beauty */
			fip = fip->next_int_user;
			if (--poll_cnt)
				goto fastloop2;
			break;
		}

		{
		    register n_unchar	mstatus REG_BX;

		    /* Has there been a polarity change on
		       some of the modem lines ?
		    */
		    if ((fip->device_flags & DF_MSI_ENABLED)
			&& ((mstatus = FAS_INB (fip, MDM_STATUS_PORT))
					& MS_ANY_DELTA))
		    {
			/* if the same modem status line
			   is responsible for a modem status
			   interrupt twice during two
			   event scheduler runs, we disable
			   modem status interrupts until we
			   process this interrupt in the event
			   scheduler
			*/
			if (mstatus & NEW_MSR & MS_ANY_DELTA)
			{
				FAS_OUTB (fip, INT_ENABLE_PORT,
					  IER &= ~IE_MODEM_STATUS);
				fip->device_flags &= ~DF_MSI_ENABLED;
				fip->device_flags |= DF_MSI_NOISE;
				++fas_msi_noise;
			}
			/* Do special RING line handling.
			   RING generates an int only on the
			   trailing edge.
			*/
			mstatus = (mstatus & ~MS_RING_PRESENT)
					| (NEW_MSR & MS_RING_PRESENT);
			if (mstatus & MS_RING_TEDGE)
				mstatus |= MS_RING_PRESENT;
			if ((mstatus ^ NEW_MSR) & MS_ANY_PRESENT)
			{
			    do_xproc = FALSE;
			    /* check hw flow flags */
			    if (!(~mstatus & fip->flow.m.oc)
					|| (~mstatus & fip->flow.m.oe)
					|| !(fip->flow_flags
						& FF_HWO_HANDSHAKE))
			    {
				if (fip->flow_flags & FF_HWO_STOPPED)
				{
				    fip->flow_flags &= ~FF_HWO_STOPPED;
				    do_xproc = TRUE;
				}
			    }
			    else
				fip->flow_flags |= FF_HWO_STOPPED;
			    /* check carrier detect
			       note: the actual carrier detect
			             processing is done in fas_mproc ()
			    */
	    		    if (!(fip->cflag & CLOCAL)
				    && (fip->flow_flags & FF_CARRIER_ON))
	    		    {
				if (!(~mstatus & fip->modem.m.ca))
				{
				    if (fip->flow_flags & FF_CARR_STOPPED)
				    {
					fip->flow_flags &= ~FF_CARR_STOPPED;
					do_xproc = TRUE;
				    }
				}
				else if (!(~NEW_MSR & fip->modem.m.ca))
				    fip->flow_flags |= FF_CARR_STOPPED;
			    }
			    if (do_xproc)
				fas_xproc (fip);
			    EVENT_SCHED (fip, EF_DO_MPROC);
			}
			else
			    EVENT_SCHED (fip, EF_RESET_DELTA_BITS);
			/* "or" the delta flags to prevent
			   excessive modem status interrupts
			*/
			NEW_MSR = mstatus | (NEW_MSR & MS_ANY_DELTA);
			was_mint = TRUE;
		    }
		}

		{
		    register n_unchar	lstatus REG_BX;

		    lstatus = FAS_INB (fip, LINE_STATUS_PORT);

		    /* Is it a receiver int ? */
		    if ((lstatus & LS_RCV_AVAIL)
				&& (fip->device_flags & DF_RDI_ENABLED))
		    {
			/* do the character processing */
			lstatus = fas_rproc (fip, lstatus);
			was_rint = TRUE;
		    }

		    /* Is it a transmitter int ? */
		    if (lstatus & LS_XMIT_AVAIL)
		    {
			if (fip->device_flags & DF_XMIT_BUSY)
			{
			    fip->device_flags &= ~DF_XMIT_BUSY;
			    /* do the character processing */
			    fas_xproc (fip);
			    if (!(fip->device_flags & DF_XMIT_BUSY))
			    {
				fip->device_flags |= DF_GUARD_TIMEOUT;
				fip->timeout_idx
				    = timeout (fas_timeout, (caddr_t) fip,
						fas_speed_ptr [BT_SELECT]
							[fip->cflag & CBAUD]
							.i.ctime);
			    }
			    was_xint = TRUE;
			}
		    }
		}

		fip = fip->next_int_user;
		poll_cnt = num_polls;	/* reset counter as there was work */
	    }
		
	    /* provide statistical infos */
	    if (was_rint)
		++sysinfo.rcvint;
	    if (was_xint)
		++sysinfo.xmtint;
	    if (was_mint)
		++sysinfo.mdmint;
	}
#if !defined (SCO) && !defined (XENIX)
	else if (fas_test_mode)
	{
		/* we are in test mode */
		fip = fas_test_fip;
		if ((FAS_FIRST_INB (fip, INT_ID_PORT) & II_CODE_MASK)
				== II_XMTD_CHAR)
		{
			fas_test_mode = FALSE;
			fas_test_vect = vect;
		}
	}
#endif

	{
		register uint	*seq_ptr REG_DI;
		register uint	port REG_BX;

		/* execute the interrupt acknowledge sequence for the
		   serial cards
		*/
		seq_ptr = &fas_int_ack_seq [0];
		port = *seq_ptr;
		if (port)
		{
			do
			{
				if (*(seq_ptr + 1) & READ_PORT)
					(void) INB (port);
				else
					(void) OUTB (port, *(seq_ptr + 1));
				port = *(seq_ptr += 2);
			} while (port);
		}
	}
}

/* Receiver interrupt handler. Translates input characters to character
   sequences as described in TERMIO(7) man page.
   (called at SPLINT)
*/
STATIC n_unchar
fas_rproc (fip, line_status)
register struct fas_internals	*fip REG_SI;
n_unchar	line_status;
{
	register unchar	*ring_ptr REG_DI;
	register n_unchar	charac REG_BX;

	/* Was there a receiver overrun ? Count them in a
	   separate counter for each UART type.
	*/
	if (line_status & LS_OVERRUN)
		++fas_overrun [DEVICE_TYPE];

	ring_ptr = fip->recv_ring_put_ptr;

	/* Translate characters from FIFO according to the TERMIO(7)
	   man page.
	*/
	do
	{
		/* ignore parity errors ? */
		if (!(fip->iflag & INPCK))
			line_status &= ~LS_PARITY_ERROR;

		charac = FAS_INB (fip, RCV_DATA_PORT);

		if (fip->iflag & ISTRIP)	/* strip off 8th bit ? */
			charac &= 0x7f;

		/* do we have some kind of character error ? */
		if (line_status & (LS_PARITY_ERROR
					| LS_FRAMING_ERROR
					| LS_BREAK_DETECTED))
		{
#if defined (HAVE_VPIX)
			if (fip->iflag & DOSMODE)
			{
			    if (fip->iflag & PARMRK)
			    {
				/* send status bits to VP/ix */
				if (fip->recv_ring_cnt < RECV_BUFF_SIZE - 2)
				{
				    fip->recv_ring_cnt += 3;
				    PUT_RECV_BUFFER (fip, ring_ptr, 0xff);
				    PUT_RECV_BUFFER (fip, ring_ptr, 1);
				    PUT_RECV_BUFFER (fip, ring_ptr,
					(line_status & (LS_PARITY_ERROR
							| LS_FRAMING_ERROR
							| LS_BREAK_DETECTED))
						| LS_RCV_AVAIL
						| LS_XMIT_AVAIL
						| LS_XMIT_COMPLETE);
				}
			    }
			    goto valid_char;
			}
			else
#endif
			/* is it a BREAK ? */
			if (line_status & LS_BREAK_DETECTED)
			{
				if (!(fip->iflag & IGNBRK))
				{
					if (fip->iflag & BRKINT)
					{
						/* do BREAK interrupt */
						EVENT_SCHED (fip, EF_DO_BRKINT);
					}
					else if (fip->iflag & PARMRK)
					{
						if (fip->recv_ring_cnt
							< RECV_BUFF_SIZE - 2)
						{
						    fip->recv_ring_cnt += 3;
						    PUT_RECV_BUFFER (fip,
								     ring_ptr,
								     0xff);
						    PUT_RECV_BUFFER (fip,
								     ring_ptr,
								     0);
						    PUT_RECV_BUFFER (fip,
								     ring_ptr,
								     0);
						}
					}
					else
					{
						if (fip->recv_ring_cnt
							< RECV_BUFF_SIZE)
						{
						    ++fip->recv_ring_cnt;
						    PUT_RECV_BUFFER (fip,
								     ring_ptr,
								     0);
						}
					}
				}
			}
			else if (!(fip->iflag & IGNPAR))
				if (fip->iflag & PARMRK)
				{
					if (fip->recv_ring_cnt
						< RECV_BUFF_SIZE - 2)
					{
						fip->recv_ring_cnt += 3;
						PUT_RECV_BUFFER (fip, ring_ptr,
								    0xff);
						PUT_RECV_BUFFER (fip, ring_ptr,
								    0);
						PUT_RECV_BUFFER (fip, ring_ptr,
								    charac);
					}
				}
				else
				{
					if (fip->recv_ring_cnt < RECV_BUFF_SIZE)
					{
						++fip->recv_ring_cnt;
						PUT_RECV_BUFFER (fip, ring_ptr,
								    0);
					}
				}
		}
		else
		{
			if (fip->iflag & IXON)
			{
				/* do output start/stop handling */
				if (fip->flow_flags & FF_SWO_STOPPED)
				{
					if ((charac == fip->start_char)
						|| (fip->iflag & IXANY))
					{
						fip->flow_flags &=
							~FF_SWO_STOPPED;
						fip->tty->t_state &= ~TTSTOP;
						/* restart output */
						fas_xproc (fip);
					}
				}
				else
				{
					if (charac == fip->stop_char)
					{
						fip->flow_flags |=
							FF_SWO_STOPPED;
						fip->tty->t_state |= TTSTOP;
					}
				}
				/* we don't put start/stop characters
				   into the receiver buffer
				*/
				if ((charac == fip->start_char)
					|| (charac == fip->stop_char))
					continue;
			}
valid_char:
			if ((charac == 0xff) && (fip->iflag & PARMRK))
			{
				if (fip->recv_ring_cnt < RECV_BUFF_SIZE - 1)
				{
					fip->recv_ring_cnt += 2;
					PUT_RECV_BUFFER (fip, ring_ptr, 0xff);
					PUT_RECV_BUFFER (fip, ring_ptr, 0xff);
				}
			}
			else
			{
				if (fip->recv_ring_cnt < RECV_BUFF_SIZE)
				{
					++fip->recv_ring_cnt;
					PUT_RECV_BUFFER (fip, ring_ptr, charac);
				}
			}
		}
	} while ((line_status = FAS_INB (fip, LINE_STATUS_PORT))
							& LS_RCV_AVAIL);

	fip->recv_ring_put_ptr = ring_ptr;

	/* schedule character transfer to CLIST buffer */
	if (!(fip->event_flags & EF_DO_RXFER)
		&& !(fip->flow_flags & FF_RXFER_STOPPED))
	{
		EVENT_SCHED (fip, EF_DO_RXFER);
	}

	/* check input buffer high water marks */
	if ((fip->recv_ring_cnt > RECV_HW_HIGH_WATER)
		&& (fip->flow_flags & (FF_HWI_HANDSHAKE | FF_HWI_STARTED))
				== (FF_HWI_HANDSHAKE | FF_HWI_STARTED))
	{
		FAS_OUTB (fip, MDM_CTL_PORT, MCR &= ~fip->flow.m.ic);
		fip->flow_flags &= ~FF_HWI_STARTED;
	}
	if ((fip->recv_ring_cnt > RECV_SW_HIGH_WATER)
			&& (fip->iflag & IXOFF)
			&& !(fip->flow_flags & FF_SWI_STOPPED))
		fas_send_xoff (fip);

	return (line_status);
}

/* Output characters to the transmitter register.
   (called at SPLINT)
*/
STATIC void
fas_xproc (fip)
register struct fas_internals	*fip REG_SI;
{
	/* proceed only if transmitter is available */
	if (fip->device_flags & (DF_XMIT_BUSY | DF_XMIT_BREAK
							| DF_XMIT_LOCKED))
		goto sched;
	if (fip->flow_flags & (FF_HWO_STOPPED | FF_CARR_STOPPED))
		goto sched;
	
	{
		register unchar	*ring_ptr REG_DI;
		register uint	num_to_output REG_BX;

		num_to_output = fip->xmit_fifo_size;
	
		/* handle XON/XOFF input flow control requests */
		if (fip->flow_flags & FF_SW_FC_REQ)
		{
			FAS_FIRST_OUTB (fip, XMT_DATA_PORT,
					(fip->flow_flags & FF_SWI_STOPPED)
						? fip->stop_char
						: fip->start_char);
			--num_to_output;
			fip->device_flags |= DF_XMIT_BUSY;
			fip->flow_flags &= ~FF_SW_FC_REQ;
			fip->tty->t_state &= ~(TTXON | TTXOFF);
			/* disable guard timeout */
			if (fip->device_flags & DF_GUARD_TIMEOUT)
			{
				fip->device_flags &= ~DF_GUARD_TIMEOUT;
				(void) untimeout (fip->timeout_idx);
			}
		}
	
		/* bail out if output is suspended by XOFF */
		if (fip->flow_flags & FF_SWO_STOPPED)
			goto sched;
	
		/* Determine how many chars to put into the transmitter
		   register.
		*/
		if (fip->xmit_ring_cnt < num_to_output)
			num_to_output = fip->xmit_ring_cnt;
	
		/* no characters available ? */
		if (!num_to_output)
			goto sched;
	
		/* output characters */
		fip->xmit_ring_cnt -= num_to_output;
	
		FAS_CTL (fip, XMT_DATA_PORT);
	
		ring_ptr = fip->xmit_ring_take_ptr;

		do
		{
			do
			{
				FAS_SAME_OUTB (fip, XMT_DATA_PORT, *ring_ptr);
				if (++ring_ptr == XMIT_BUFF_END)
					break;
			} while (--num_to_output);
			if (!num_to_output)
				break;
			ring_ptr = XMIT_BUFF_BEGIN;
		} while (--num_to_output);
	
		fip->xmit_ring_take_ptr = ring_ptr;

		/* signal that transmitter is busy now */
		fip->device_flags |= DF_XMIT_BUSY;
		/* disable guard timeout */
		if (fip->device_flags & DF_GUARD_TIMEOUT)
		{
			fip->device_flags &= ~DF_GUARD_TIMEOUT;
			(void) untimeout (fip->timeout_idx);
		}
	}

	/* schedule fas_xxfer () if there are more characters to transfer
	   into the transmitter ring buffer
	*/
sched:
	{
		register struct tty	*ttyp REG_DI;

		if (!(fip->event_flags & EF_DO_XXFER)
			&& (fip->xmit_ring_max > fip->xmit_ring_cnt)
			&& ((ttyp = fip->tty)->t_outq.c_cc
				|| (ttyp->t_tbuf.c_ptr
					&& ttyp->t_tbuf.c_count)))
		{
			EVENT_SCHED (fip, EF_DO_XXFER);
		}
	}
}

/* Open device physically.
   (called at SPLINT)
*/
STATIC void
fas_open_device (fip)
register struct fas_internals	*fip REG_SI;
{
	/* init some variables */
	fip->device_flags &= DF_DEVICE_CONFIGURED | DF_DEVICE_LOCKED
				| DF_CTL_FIRST | DF_CTL_EVERY | DF_HUP_PROTECT
				| DF_NO_OVERRUN;
	fip->device_flags |= DF_DEVICE_OPEN;
	fip->flow_flags &= FF_NEW_CTSRTS;
	fip->flow_flags |= FF_CD_ENABLED;
	fip->event_flags = 0;
	fip->recv_ring_put_ptr = fip->recv_ring_take_ptr
				= RECV_BUFF_BEGIN;
	fip->recv_ring_cnt = 0;
	fip->xmit_ring_put_ptr = fip->xmit_ring_take_ptr
				= XMIT_BUFF_BEGIN;
	fip->xmit_ring_cnt = 0;
	fip->rxfer_timeout = MAX_RXFER_DELAY / EVENT_TIME;
	LCR = 0;
	MCR &= ~(fip->modem.m.di | fip->modem.m.ei | fip->modem.m.eo);
	fip->start_char = CSTART;
	fip->stop_char = CSTOP;
#if defined (HAVE_VPIX)
	/* initialize VP/ix related variables */
	fip->v86_proc = (v86_t *) NULL;
	fip->v86_intmask = 0;
#endif

	/* hook into the interrupt users chain */
	if ((fip->next_int_user = fas_first_int_user))
		fip->next_int_user->prev_int_user = fip;
	else
		fas_last_int_user = fip;
	fas_first_int_user = fip;
	fip->prev_int_user = (struct fas_internals *) NULL;

	/* enable FIFOs */
	if (DEVICE_TYPE == TYPE_NS16550A)
		FAS_FIRST_OUTB (fip, NS_FIFO_CTL_PORT, FCR | NS_FIFO_CLR_XMIT);
	else if (DEVICE_TYPE == TYPE_I82510)
	{
		FAS_FIRST_OUTB (fip, I_BANK_PORT, I_BANK_2);
		FAS_OUTB (fip, I_IDM_PORT, I_FIFO_SETUP_CMD);
		FAS_OUTB (fip, I_BANK_PORT, I_BANK_1);
		FAS_OUTB (fip, I_TCM_PORT, I_FIFO_CLR_XMIT);
		FAS_OUTB (fip, I_BANK_PORT, I_BANK_0);
	}

	/* enable and clear transmitter interrupts */
	FAS_FIRST_OUTB (fip, INT_ENABLE_PORT,
				IER = IE_XMIT_HOLDING_BUFFER_EMPTY);
	(void) FAS_INB (fip, INT_ID_PORT);

	fas_param (fip, HARD_INIT);	/* set up port regs */
}

/* Close device physically.
   (called at SPLINT)
*/
STATIC void
fas_close_device (fip)
register struct fas_internals	*fip REG_SI;
{
	/* disable UART interrupts */
	FAS_FIRST_OUTB (fip, INT_ENABLE_PORT, IER = IE_NONE);

	/* reset break level */
	FAS_OUTB (fip, LINE_CTL_PORT, LCR = 0);

	/* clear some variables */
	fip->device_flags &= ~DF_DEVICE_OPEN;
	fip->event_flags = 0;

	/* unhook from interrupt users chain */
	if (fip->prev_int_user)
		fip->prev_int_user->next_int_user = fip->next_int_user;
	else
		fas_first_int_user = fip->next_int_user;
	if (fip->next_int_user)
		fip->next_int_user->prev_int_user = fip->prev_int_user;
	else
		fas_last_int_user = fip->prev_int_user;
}

/* Set up a port according to the given termio structure.
   (called at SPLINT)
*/
STATIC void
fas_param (fip, init_type)
register struct	fas_internals	*fip REG_SI;
int	init_type;
{
	register struct tty	*ttyp REG_DI;
	bool	do_flush;
	int	old_level;

	ttyp = fip->tty;

	{
	    register uint	cflag REG_BX;

	    cflag = ttyp->t_cflag;
	    fip->iflag = ttyp->t_iflag;
	    do_flush = FALSE;

#if defined (HAVE_VPIX)
	    /* we don't set port registers if we are in dos mode */
	    if (fip->iflag & DOSMODE)
	    {
		/* This is a kludge. We don't know what baud rate
		   DOS will use. Therefore, we assume a rather low
		   one to be on the safe side.
		*/
		cflag = (cflag & ~CBAUD) | B300;
		goto setflags;
	    }
#endif
	    /* Make sure that we have a valid baud rate. If we don't
	       get one, take the previous baud rate.
	    */
	    if ((cflag & CBAUD) == B0)
		cflag = (cflag & ~CBAUD) | (fip->cflag & CBAUD);

	    /* if soft init mode: don't set port registers if cflag
	       didn't change
	    */
	    if ((init_type == SOFT_INIT)
			&& !((cflag ^ fip->cflag)
				& (CBAUD | CSIZE | CSTOPB | PARENB | PARODD)))
		goto setflags;

	    /* don't change break flag */
	    LCR &= LC_SET_BREAK_LEVEL;

	    /* set character size */
	    switch (cflag & CSIZE)
	    {
		case CS5:
			LCR |= LC_WORDLEN_5;
			break;

		case CS6:
			LCR |= LC_WORDLEN_6;
			break;

		case CS7:
			LCR |= LC_WORDLEN_7;
			break;

		default:
			LCR |= LC_WORDLEN_8;
			break;
	    }

	    /* set # of stop bits */
	    if (cflag & CSTOPB)
		LCR |= LC_STOPBITS_LONG;

	    /* set parity */
	    if (cflag & PARENB)
	    {
		LCR |= LC_ENABLE_PARITY;

		if (!(cflag & PARODD))
			LCR |= LC_EVEN_PARITY;
	    }

	    /* set LCR and baud rate */
	    FAS_FIRST_OUTB (fip, LINE_CTL_PORT, LCR | LC_ENABLE_DIVISOR);
	    FAS_OUTB (fip, DIVISOR_LSB_PORT, fas_speed_ptr [BT_SELECT]
								[cflag & CBAUD]
								.div.b.low);
	    FAS_OUTB (fip, DIVISOR_MSB_PORT, fas_speed_ptr [BT_SELECT]
								[cflag & CBAUD]
								.div.b.high);
	    FAS_OUTB (fip, LINE_CTL_PORT, LCR);

setflags:
	    /* set dynamic xmit ring buffer size */
	    fip->xmit_ring_max = fas_speed_ptr [BT_SELECT] [cflag & CBAUD]
							.xbuf_size;

	    /* disable modem control signals if required by open mode */
	    if (fip->o_state & OS_CLOCAL)
		cflag |= CLOCAL;

	    /* Select hardware handshake depending on the minor device
	       number and various termio flags (if they are available).
	    */
	    fip->flow_flags &= ~(FF_HWO_HANDSHAKE
					| FF_HWI_HANDSHAKE
					| FF_HDX_HANDSHAKE);
	    if (fip->o_state & (OS_HWO_HANDSHAKE | OS_HWI_HANDSHAKE
					| OS_HDX_HANDSHAKE))
	    {
		if (fip->o_state & OS_HWO_HANDSHAKE)
			fip->flow_flags |= FF_HWO_HANDSHAKE;
		if (fip->o_state & OS_HWI_HANDSHAKE)
			fip->flow_flags |= FF_HWI_HANDSHAKE;
		if (fip->o_state & OS_HDX_HANDSHAKE)
			fip->flow_flags |= FF_HDX_HANDSHAKE;
	    }
	    else
	    {
#if defined (CTSXON) && defined (RTSXOFF)	/* SVR4 compatibility */
		if (fip->iflag & CTSXON)
			fip->flow_flags |= FF_HWO_HANDSHAKE;
		if (fip->iflag & RTSXOFF)
			fip->flow_flags |= FF_HWI_HANDSHAKE;
#else
#if defined (CTSFLOW) && defined (RTSFLOW)	/* SCO Xenix compatibility */
		if (!(cflag & CLOCAL) || (fip->flow_flags & FF_NEW_CTSRTS))
		{
#if defined (ORTSFL)	/* SCO UNIX 3.2.4.2 compatibility */
			if ((cflag & (ORTSFL | CTSFLOW | RTSFLOW)) == ORTSFL)
				fip->flow_flags |= FF_HWO_HANDSHAKE
							| FF_HWI_HANDSHAKE;
			else
			{
				if (cflag & CTSFLOW)
					fip->flow_flags |= FF_HWO_HANDSHAKE;
				if (cflag & RTSFLOW)
					fip->flow_flags |= ((fip->flow_flags
								& FF_NEW_CTSRTS)
							   || !(cflag & ORTSFL))
							    ? FF_HWI_HANDSHAKE
							    : FF_HDX_HANDSHAKE;
			}
#else
#if defined (CRTSFL)	/* SCO UNIX 3.2.4 compatibility */
			if ((cflag & (CRTSFL | CTSFLOW | RTSFLOW)) == CRTSFL)
				fip->flow_flags |= FF_HWO_HANDSHAKE
							| FF_HWI_HANDSHAKE;
			else
#endif
			{
				if (cflag & CTSFLOW)
					fip->flow_flags |= FF_HWO_HANDSHAKE;
				if (cflag & RTSFLOW)
					fip->flow_flags |= (fip->flow_flags
								& FF_NEW_CTSRTS)
							    ? FF_HWI_HANDSHAKE
							    : FF_HDX_HANDSHAKE;
			}
#endif
		}
#endif
#endif
	    }

	    /* Determine whether to enable MSI or not.
	       Set the interrupt enable port accordingly.
	    */
#if defined (HAVE_VPIX)
	    if ((cflag & CLOCAL) && !(fip->flow_flags & FF_HWO_HANDSHAKE)
			&& !(fip->iflag & DOSMODE))
#else
	    if ((cflag & CLOCAL) && !(fip->flow_flags & FF_HWO_HANDSHAKE))
#endif
	    {
		if (fip->device_flags & DF_MSI_ENABLED)
		{
			FAS_FIRST_OUTB (fip, INT_ENABLE_PORT,
					IER &= ~IE_MODEM_STATUS);
			fip->device_flags &= ~DF_MSI_ENABLED;
		}
		fip->device_flags &= ~DF_MSI_NOISE;
	    }
	    else
	    {
		if (!(fip->device_flags & (DF_MSI_ENABLED | DF_MSI_NOISE)))
		{
			/* read current modem status lines */
			MSR = NEW_MSR = FAS_FIRST_INB (fip, MDM_STATUS_PORT)
						& (MS_CTS_PRESENT
							| MS_DSR_PRESENT
							| MS_DCD_PRESENT);
			FAS_OUTB (fip, INT_ENABLE_PORT, IER |= IE_MODEM_STATUS);
			fip->device_flags |= DF_MSI_ENABLED;
		}
	    }

	    /* Fake the carrier detect state flag if CLOCAL mode or if
	       requested by open mode.
	    */
	    if ((cflag & CLOCAL)
			|| (fip->flow_flags & FF_CD_ENABLED
				&& !(~MSR & fip->modem.m.ca))
			|| ((fip->o_state & OS_FAKE_CARRIER_ON)
				&& ((fip->flow_flags & FF_CARRIER_ON)
					|| !(fip->o_state & OS_OPEN_STATES))))
	    {
		fip->flow_flags |= FF_CD_ENABLED | FF_CARRIER_ON;
		ttyp->t_state |= CARR_ON;
	    }
	    else
	    {
		if (fip->flow_flags & FF_CARRIER_ON)
		{
			fip->flow_flags &= ((fip->device_flags
							& DF_HUP_PROTECT)
						    && (fip->o_state
							& OS_OPEN_STATES))
						? ~(FF_CD_ENABLED
							| FF_CARRIER_ON)
						: ~FF_CARRIER_ON;
			/* flush buffers on carrier drop */
			do_flush = TRUE;
		}
		ttyp->t_state &= ~CARR_ON;
	    }

	    if ((fip->flow_flags & FF_CARRIER_ON) && (cflag & CREAD))
	    {
		/* enable receiver data interrupts */
		if (!(fip->device_flags & DF_RDI_ENABLED))
			fas_rdi_enable (fip);
	    }
	    else
	    {
		/* disable receiver data interrupts */
		if (fip->device_flags & DF_RDI_ENABLED)
		{
			FAS_FIRST_OUTB (fip, INT_ENABLE_PORT,
					IER &= ~IE_RECV_DATA_AVAILABLE);
			fip->device_flags &= ~DF_RDI_ENABLED;
		}
	    }

	    fip->cflag = cflag;
	}

#if defined (XCLUDE)	/* SYSV 3.2 Xenix compatibility */
	/* Permit exclusive use of this device. */
	if (ttyp->t_lflag & XCLUDE)
		fip->o_state |= OS_EXCLUSIVE_OPEN_2;
	else
		fip->o_state &= ~OS_EXCLUSIVE_OPEN_2;
#endif

	/* re-link fas_internals structure */
	fas_pos_by_speed (fip);

	/* flush buffers if necessary */
	if (do_flush)
	{
		/* lock transmitter */
		fip->device_flags |= DF_XMIT_LOCKED;
		old_level = SPLWRK ();
		(void) ttyflush (ttyp, FREAD | FWRITE);
		(void) splx (old_level);
		/* enable transmitter */
		fip->device_flags &= ~DF_XMIT_LOCKED;
	}

	/* setup handshake flags */
#if defined (HAVE_VPIX)
	if ((fip->flow_flags & (FF_HWI_HANDSHAKE | FF_HDX_HANDSHAKE))
		|| !(fip->iflag & DOSMODE))
#endif
	{
		register n_unchar	mcr REG_BX;

		/* clear flow control output bits */
		mcr = MCR & ~(fip->flow.m.ic | fip->flow.m.hc);
		fip->flow_flags &= ~(FF_HWI_STARTED | FF_HDX_STARTED);

		/* raise flow control bits if necessary */
		if (fip->flow_flags & FF_HWI_HANDSHAKE)
		{
			if (fip->recv_ring_cnt < RECV_HW_LOW_WATER)
			{
				mcr |= fip->flow.m.ic;
				fip->flow_flags |= FF_HWI_STARTED;
			}
		}
		else if (fip->flow_flags & FF_HDX_HANDSHAKE)
		{
			if (fip->flow_flags & FF_OUTPUT_BUSY)
			{
				mcr |= fip->flow.m.hc;
				fip->flow_flags |= FF_HDX_STARTED;
			}
		}
		else if (!(fip->flow_flags & FF_DEF_HHO_LOW))
			mcr |= fip->flow.m.hc;

		/* set modem enable bits */
#if defined (HAVE_VPIX)
		if (!(fip->iflag & DOSMODE))
#endif
		{
			if ((ttyp->t_cflag & CBAUD) != B0
				&& (fip->flow_flags & FF_CD_ENABLED))
			{
				mcr |= (fip->o_state & OS_WAIT_OPEN)
						? fip->modem.m.ei
						: fip->modem.m.eo;
			}
			else
			{
				mcr &= (fip->o_state & OS_WAIT_OPEN)
						? ~fip->modem.m.ei
						: ~fip->modem.m.eo;
			}
		}

		if ((init_type != SOFT_INIT) || (mcr != MCR))
			FAS_FIRST_OUTB (fip, MDM_CTL_PORT, MCR = mcr);
	}

	/* set hardware output flow control flags */
	if (!(~NEW_MSR & fip->flow.m.oc)
			|| (~NEW_MSR & fip->flow.m.oe)
			|| !(fip->flow_flags & FF_HWO_HANDSHAKE))
		fip->flow_flags &= ~FF_HWO_STOPPED;
	else
		fip->flow_flags |= FF_HWO_STOPPED;

	/* check software input flow control */
	if (fip->flow_flags & FF_SWI_STOPPED)
	{
		if (!(fip->iflag & IXOFF)
				|| (fip->recv_ring_cnt < RECV_SW_LOW_WATER))
			fas_send_xon (fip);
	}
	else
	{
		if ((fip->iflag & IXOFF)
				&& (fip->recv_ring_cnt > RECV_SW_HIGH_WATER))
			fas_send_xoff (fip);
	}

	/* restart output */
	fas_xproc (fip);
}

/* Re-link fas_internals structure to interrupt users chain according
   to the baud rate (fastes port comes first).
   (called at SPLINT)
*/
STATIC void
fas_pos_by_speed (fipp)
struct fas_internals	*fipp;
{
	register struct fas_internals	*fip REG_SI;

	{
	    register n_unchar	my_prio REG_BX;
	    register n_unchar	prio REG_CX;
	    ulong	my_speed;

	    fip = fipp;

	    /* compute our own speed */
	    if (fip->flow_flags & FF_CARRIER_ON)
	    {
		INT_PRIO = my_prio = (fip->device_flags & DF_NO_OVERRUN)
					? NUMBER_OF_TYPES
					: DEVICE_TYPE;
		my_speed = fas_baud_ptr [BT_SELECT] [fip->cflag & CBAUD];
	    }
	    else
	    {
		my_prio = NUMBER_OF_TYPES + 1;
		INT_PRIO = NUMBER_OF_TYPES + 1;
		my_speed = 0L;
	    }

	    /* unhook us from the interrupt users chain */
	    if (fip->prev_int_user)
		fip->prev_int_user->next_int_user = fip->next_int_user;
	    else
		fas_first_int_user = fip->next_int_user;
	    if (fip->next_int_user)
		fip->next_int_user->prev_int_user = fip->prev_int_user;
	    else
		fas_last_int_user = fip->prev_int_user;

	    /* find a new position for us */
	    for (fip = fas_first_int_user; fip; fip = fip->next_int_user)
	    {
fastloop4:
		if (my_prio > (prio = INT_PRIO))
		{
			/* speed beats beauty */
			fip = fip->next_int_user;
			if (fip)
				goto fastloop4;
			break;
		}

		if (my_prio < prio)
			break;	/* found a new position */
		if (my_speed >= ((prio == NUMBER_OF_TYPES + 1)
					? 0L
					: fas_baud_ptr [BT_SELECT]
						[fip->cflag & CBAUD]))
			break;	/* found a new position */
	    }
	}

	{
	    register struct fas_internals	*my_fip REG_DI;

	    my_fip = fipp;

	    /* now hook us into the new position */
	    if (fip)
	    {
		/* we are in front of another entry */
		if ((my_fip->prev_int_user = fip->prev_int_user))
			my_fip->prev_int_user->next_int_user = my_fip;
		else
			fas_first_int_user = my_fip;
		fip->prev_int_user = my_fip;
		my_fip->next_int_user = fip;
	    }
	    else
	    {
		/* we are the last entry */
		if ((my_fip->prev_int_user = fas_last_int_user))
			my_fip->prev_int_user->next_int_user = my_fip;
		else
			fas_first_int_user = my_fip;
		fas_last_int_user = my_fip;
		my_fip->next_int_user = (struct fas_internals *) NULL;
	    }
	}
}

/* Asynchronous event handler. Scheduled by functions that can't do the
   processing themselves because of execution time restrictions.
   (called via timeout () at an OS dependent SPL)
*/
STATIC void
fas_event (dummy)
void	*dummy;
{
	register struct fas_internals	*fip REG_SI;
	register struct tty	*ttyp REG_DI;
	register uint	count REG_BX;
	bool	idle;
	int	old_level;

	/* allow pending tty interrupts */
	old_level = SPLWRK ();
	(void) SPLINT ();

	idle = TRUE;

	/* loop through all fas_internals structures */
	for (fip = &fas_internals [0], count = fas_physical_units; count;
			++fip, --count)
	{
fastloop3:
		/* process only structures that actually need service */
		if (!fip->event_flags)
		{
			/* speed beats beauty */
			++fip;
			if (--count)
				goto fastloop3;
			break;
		}

		/* we have something to do */
		idle = FALSE;
		ttyp = fip->tty;

		/* check the modem signals */
		if (fip->event_flags & (EF_DO_MPROC | EF_RESET_DELTA_BITS))
		{
			NEW_MSR &= ~MS_ANY_DELTA;
			if (fip->event_flags & EF_DO_MPROC)
			{
				fip->event_flags &= ~(EF_DO_MPROC
							| EF_RESET_DELTA_BITS);
				fas_mproc (fip);
			}
			else
				fip->event_flags &= ~EF_RESET_DELTA_BITS;

			/* re-enable modem status interrupts if appropriate */
			if (fip->device_flags & DF_MSI_NOISE)
			{
				FAS_FIRST_OUTB (fip, INT_ENABLE_PORT,
							IER |= IE_MODEM_STATUS);
				fip->device_flags |= DF_MSI_ENABLED;
				fip->device_flags &= ~DF_MSI_NOISE;
			}
		}

		/* do the break interrupt */
		if (fip->event_flags & EF_DO_BRKINT)
		{
			fip->event_flags &= ~EF_DO_BRKINT;
			if (fip->flow_flags & FF_CARRIER_ON)
			{
				(void) SPLWRK ();
				(void) (*linesw [(unchar) ttyp->t_line].l_input)
								(ttyp, L_BREAK);
				(void) SPLINT ();
			}
		}

		/* transfer characters to the CLIST input buffer */
		if (fip->event_flags & EF_DO_RXFER)
		{
#if defined (HAVE_VPIX)
			fip->event_flags &= ~(EF_DO_RXFER | EF_WAKEUP_VPIX);
#else
			fip->event_flags &= ~EF_DO_RXFER;
#endif
			if ((fip->flow_flags & (FF_CARRIER_ON
							| FF_RXFER_STOPPED))
				== FF_CARRIER_ON)
			{
			    /* if not done, re-schedule this event */
			    if (!fas_rxfer (fip))
				fip->event_flags |= EF_DO_RXFER;
			    /* check input buffer low water marks */
			    if ((fip->recv_ring_cnt < RECV_HW_LOW_WATER)
				    && (fip->flow_flags & (FF_HWI_HANDSHAKE
							    | FF_HWI_STARTED))
					== FF_HWI_HANDSHAKE)
			    {
				FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR |= fip->flow.m.ic);
				fip->flow_flags |= FF_HWI_STARTED;
			    }
			    if ((fip->recv_ring_cnt < RECV_SW_LOW_WATER)
				    && (fip->iflag & IXOFF)
				    && (fip->flow_flags & FF_SWI_STOPPED))
				fas_send_xon (fip);
			}
#if defined (HAVE_VPIX)
			/* send pseudorupt to VP/ix */
			if ((fip->iflag & DOSMODE) && fip->v86_proc
				&& (ttyp->t_rawq.c_cc || ttyp->t_canq.c_cc))
			{
				(void) SPLWRK ();
				(void) v86setint (fip->v86_proc,
							fip->v86_intmask);
				(void) SPLINT ();
				fip->event_flags |= EF_WAKEUP_VPIX;
			}
#endif
		}
#if defined (HAVE_VPIX)
		else if (fip->event_flags & EF_WAKEUP_VPIX)
		{
			/* send pseudorupt to VP/ix */
			if ((fip->iflag & DOSMODE) && fip->v86_proc
				&& (ttyp->t_rawq.c_cc || ttyp->t_canq.c_cc))
			{
				(void) SPLWRK ();
				(void) v86setint (fip->v86_proc,
							fip->v86_intmask);
				(void) SPLINT ();
			}
			else
				fip->event_flags &= ~EF_WAKEUP_VPIX;
		}
#endif

		/* transfer characters to the output ring buffer */
		if (fip->event_flags & EF_DO_XXFER)
		{
			fip->event_flags &= ~EF_DO_XXFER;
			if (fip->flow_flags & FF_CARRIER_ON)
			{
				fas_xxfer (fip);
				if (fip->flow_flags & FF_OUTPUT_BUSY)
					ttyp->t_state |= BUSY;
				/* check half duplex output flow control */
				if ((fip->flow_flags & (FF_HDX_HANDSHAKE
							| FF_HDX_STARTED
							| FF_OUTPUT_BUSY))
						== (FF_HDX_HANDSHAKE
							| FF_OUTPUT_BUSY))
				{
				    FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						    MCR |= fip->flow.m.hc);
				    fip->flow_flags |= FF_HDX_STARTED;
				}
				/* output characters */
				fas_xproc (fip);
			}
			else
			{
			    /* throw away characters in output queue */
			    (void) SPLWRK ();
			    (void) ttyflush (ttyp, FWRITE);
			    (void) SPLINT ();
			}
		}

		/* allow pending tty interrupts */
		(void) SPLWRK ();
		(void) SPLINT ();
	}

	event_scheduled = FALSE;

	/* check whether there was something to do */
	if (!idle)
	{
		/* there was work to do, so it's likely that there
		   also will be something to do in a short time from
		   now, so we schedule the next event processing
		*/
		event_scheduled = TRUE;
		(void) timeout (fas_event, (caddr_t) NULL,
				(EVENT_TIME) * (HZ) / 1000);
	}

	(void) splx (old_level);
}

/* Modem status handler.
   (called at SPLINT)
*/
STATIC void
fas_mproc (fip)
register struct fas_internals	*fip REG_SI;
{
	register n_unchar	mdm_status REG_BX;
	int	old_level;

	mdm_status = NEW_MSR;
	NEW_MSR &= ~MS_RING_PRESENT;

	/* Check the carrier detect signal and set the state flags
	   accordingly. Also, if not in clocal mode, send SIGHUP on
	   carrier loss and flush the buffers.
	*/
	if (!(fip->cflag & CLOCAL) && (fip->flow_flags & FF_CD_ENABLED))
	{
		register struct tty	*ttyp REG_DI;

		ttyp = fip->tty;
		if (!(~mdm_status & fip->modem.m.ca))
		{
			/* Wake up dialin process on carrier low->high. */
			if (!(fip->flow_flags & FF_CARRIER_ON))
			{
				fip->flow_flags |= FF_CARRIER_ON;
				if ((ttyp->t_state |= CARR_ON) & WOPEN)
					(void) wakeup ((caddr_t) &ttyp->t_canq);
				/* enable receiver data interrupts */
				if ((fip->cflag & CREAD)
						&& !(fip->device_flags
							& DF_RDI_ENABLED))
					fas_rdi_enable (fip);
				/* re-link fas_internals structure */
				fas_pos_by_speed (fip);
			}
		}
		else
		{
			if (!(~MSR & fip->modem.m.ca))
			{
				if ((fip->device_flags & DF_HUP_PROTECT)
					&& (fip->o_state & OS_OPEN_STATES))
				{
				    FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						    MCR &= (fip->o_state
								& OS_WAIT_OPEN)
							    ? ~fip->modem.m.ei
							    : ~fip->modem.m.eo);
				    fip->flow_flags &= ~(FF_CD_ENABLED
							    | FF_CARRIER_ON);
				}
				else
				    fip->flow_flags &= ~FF_CARRIER_ON;
				ttyp->t_state &= ~CARR_ON;
				/* disable receiver data interrupts */
				if (fip->device_flags & DF_RDI_ENABLED)
				{
				    FAS_FIRST_OUTB (fip, INT_ENABLE_PORT,
						IER &= ~IE_RECV_DATA_AVAILABLE);
				    fip->device_flags &= ~DF_RDI_ENABLED;
				}
				/* re-link fas_internals structure */
				fas_pos_by_speed (fip);
				old_level = SPLWRK ();
				if (ttyp->t_pgrp)
				{
					(void) signal (ttyp->t_pgrp, SIGHUP);
#if defined (SIGCONT)
					/* make sure processes are awake */
					(void) signal (ttyp->t_pgrp, SIGCONT);
#endif
				}
				(void) ttyflush (ttyp, FREAD | FWRITE);
				(void) splx (old_level);
			}
		}
	}

#if defined (HAVE_VPIX)
	if ((fip->iflag & (DOSMODE | PARMRK)) == (DOSMODE | PARMRK)
		&& (fip->flow_flags & FF_CARRIER_ON)
		&& fip->recv_ring_cnt < RECV_BUFF_SIZE - 2)
	{
		register n_unchar	vpix_status REG_CX;

		/* prepare status bits for VP/ix */
		vpix_status = (((mdm_status ^ MSR) >> 4) & MS_ANY_DELTA)
				| (mdm_status & (MS_CTS_PRESENT
							| MS_DSR_PRESENT
							| MS_DCD_PRESENT));
		if (fip->flow_flags & FF_HWO_HANDSHAKE)
		{
			vpix_status &= ~((n_unchar) (fip->flow.m.oc
							| fip->flow.m.oe) >> 4);
			vpix_status |= fip->flow.m.oc | fip->flow.m.oe;
		}
		if (!(fip->cflag & CLOCAL))
		{
			vpix_status &= ~((n_unchar) fip->modem.m.ca >> 4);
			vpix_status |= fip->modem.m.ca;
		}
		/* send status bits to VP/ix */
		if (vpix_status & MS_ANY_DELTA)
		{
			fip->recv_ring_cnt += 3;
			PUT_RECV_BUFFER (fip, fip->recv_ring_put_ptr, 0xff);
			PUT_RECV_BUFFER (fip, fip->recv_ring_put_ptr, 2);
			PUT_RECV_BUFFER (fip, fip->recv_ring_put_ptr,
						vpix_status);
			EVENT_SCHED (fip, EF_DO_RXFER);
			/* check input buffer high water marks */
			if ((fip->recv_ring_cnt > RECV_HW_HIGH_WATER)
				    && (fip->flow_flags & (FF_HWI_HANDSHAKE
							    | FF_HWI_STARTED))
					== (FF_HWI_HANDSHAKE | FF_HWI_STARTED))
			{
				FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR &= ~fip->flow.m.ic);
				fip->flow_flags &= ~FF_HWI_STARTED;
			}
			if ((fip->recv_ring_cnt > RECV_SW_HIGH_WATER)
				    && (fip->iflag & IXOFF)
				    && !(fip->flow_flags & FF_SWI_STOPPED))
				fas_send_xoff (fip);
		}
	}
#endif
	MSR = mdm_status & ~MS_RING_PRESENT;

	/* re-enable transmitter if it was stopped by carrier detect */
	if (fip->flow_flags & FF_CARR_STOPPED)
	{
		fip->flow_flags &= ~FF_CARR_STOPPED;
		fas_xproc (fip);
	}
}

/* Receiver ring buffer -> CLIST buffer transfer function.
   (called at SPLINT)
*/
STATIC bool
fas_rxfer (fipp)
struct fas_internals	*fipp;
{
	/* Defining register variables with function-wide scope would
	   break the FAS_MEM_COPY () macro below. So don't do that!
	*/
	unchar	*tmp_srcptr, *tmp_destptr;
	uint	num_save;
	uint	recv_ring_cnt;
	int	old_level;

	num_save = 0;

	for (;;)
	{
	    /* determine how many characters we have to copy */
	    {
		register struct fas_internals	*fip REG_SI;
		register struct tty	*ttyp REG_DI;
		register uint	num_to_xfer REG_BX;

		old_level = SPLWRK ();
		fip = fipp;
		ttyp = fip->tty;

		/* if the target buffer is full, or if there is no
		   target buffer available, try to get a new one,
		   otherwise return
		*/
		if (num_save || !ttyp->t_rbuf.c_ptr)
		{
			(void) (*linesw [(unchar) ttyp->t_line].l_input)
								(ttyp, L_BUF);
			if (!ttyp->t_rbuf.c_ptr)
			{
				(void) splx (old_level);
				return (FALSE);
			}
		}

		/* return if there aren't any characters to transfer */
		if (!(recv_ring_cnt = fip->recv_ring_cnt))
		{
			(void) splx (old_level);
			return (TRUE);
		}

		/* determine how many characters to transfer */
#if defined (XENIX)
		if (ttyp->t_lflag & ICANON)
#else
		if ((ttyp->t_lflag & ICANON) || ttyp->t_term)
#endif
		{
			num_to_xfer = recv_ring_cnt;
		}
		else
		{
			num_to_xfer = max_rawq_count - (uint) ttyp->t_rawq.c_cc;
			if (!num_to_xfer)
			{
				(void) splx (old_level);
				return (FALSE);	/* input buffer full */
			}
			if (recv_ring_cnt < num_to_xfer)
				num_to_xfer = recv_ring_cnt;
			/* if there are too few remaining characters to
			   completely fill a CLIST buffer, postpone the
			   transfer of these characters for a number of
			   event cycles on the assumption that by then
			   more characters have arrived
			*/
			if (ttyp->t_rbuf.c_count > num_to_xfer
				&& (num_save || --fip->rxfer_timeout))
			{
				(void) splx (old_level);
				return (FALSE);	/* input buffer full */
			}
		}

		/* reset timeout counter */
		fip->rxfer_timeout = MAX_RXFER_DELAY / EVENT_TIME;

		/* determine how many characters are in one contiguous block */
		if (ttyp->t_rbuf.c_count < num_to_xfer)
			num_to_xfer = ttyp->t_rbuf.c_count;
		if ((uint) (RECV_BUFF_END - fip->recv_ring_take_ptr)
				< num_to_xfer)
			num_to_xfer = RECV_BUFF_END - fip->recv_ring_take_ptr;

		if (!num_to_xfer)
		{
			/* invalid count, shouldn't happen */
			(void) splx (old_level);
			return (FALSE);
		}

		tmp_srcptr = fip->recv_ring_take_ptr;
		if ((fip->recv_ring_take_ptr += num_to_xfer) == RECV_BUFF_END)
			fip->recv_ring_take_ptr = RECV_BUFF_BEGIN;
		tmp_destptr = (unchar *) ttyp->t_rbuf.c_ptr;
		ttyp->t_rbuf.c_count -= num_to_xfer;
		num_save = num_to_xfer;
	    }

	    /* now copy from the ring buffer to the clist buffer */
	    FAS_MEM_COPY (tmp_destptr, tmp_srcptr, num_save);

	    (void) splx (old_level);
	    fipp->recv_ring_cnt -= num_save;
	}
}

/* CLIST buffer -> transmitter ring buffer transfer function.
   (called at SPLINT)
*/
STATIC void
fas_xxfer (fipp)
struct fas_internals	*fipp;
{
	/* Defining register variables with function-wide scope would
	   break the FAS_MEM_COPY () macro below. So don't do that!
	*/
	unchar	*tmp_srcptr, *tmp_destptr;
	uint	num_save;
	uint	xmit_ring_cnt;
	int	old_level;

	for (;;)
	{
	    /* determine how many characters we have to copy */
	    {
		register struct fas_internals	*fip REG_SI;
		register struct tty	*ttyp REG_DI;
		register uint	num_to_xfer REG_BX;

		old_level = SPLWRK ();
		fip = fipp;
		ttyp = fip->tty;

		/* Check if tbuf is empty. If it is empty, reset buffer
		   pointer and counter and get the next chunk of output
		   characters.
		*/
		if (!ttyp->t_tbuf.c_ptr || !ttyp->t_tbuf.c_count)
		{
			if (ttyp->t_tbuf.c_ptr)
				ttyp->t_tbuf.c_ptr -= ttyp->t_tbuf.c_size;
			if (!((*linesw [(unchar) ttyp->t_line].l_output) (ttyp)
					& CPRES))
			{
				(void) splx (old_level);
				return;
			}
		}

		/* Return if transmitter ring buffer is full. */
		if (fip->xmit_ring_max <= (xmit_ring_cnt = fip->xmit_ring_cnt))
		{
			(void) splx (old_level);
			return;
		}

		/* set the maximum character limit */
		num_to_xfer = fip->xmit_ring_max - xmit_ring_cnt;

		/* Determine how many chars to transfer this time. */
		if (ttyp->t_tbuf.c_count < num_to_xfer)
			num_to_xfer = ttyp->t_tbuf.c_count;
		if ((uint) (XMIT_BUFF_END - fip->xmit_ring_put_ptr)
				< num_to_xfer)
			num_to_xfer = XMIT_BUFF_END - fip->xmit_ring_put_ptr;
		if (!num_to_xfer)
		{
			/* invalid count, shouldn't happen */
			(void) splx (old_level);
			return;
		}

		fip->flow_flags |= FF_OUTPUT_BUSY;
		tmp_srcptr = (unchar *) ttyp->t_tbuf.c_ptr;
		ttyp->t_tbuf.c_ptr += num_to_xfer;
		ttyp->t_tbuf.c_count -= num_to_xfer;
		tmp_destptr = fip->xmit_ring_put_ptr;
		if ((fip->xmit_ring_put_ptr += num_to_xfer) == XMIT_BUFF_END)
			fip->xmit_ring_put_ptr = XMIT_BUFF_BEGIN;
		num_save = num_to_xfer;
	    }

	    /* now copy from the clist buffer to the ring buffer */
	    FAS_MEM_COPY (tmp_destptr, tmp_srcptr, num_save);

	    (void) splx (old_level);
	    fipp->xmit_ring_cnt += num_save;
	}
}

/* Handle hangup after last close.
   (called via timeout () at an OS dependent SPL)
*/
STATIC void
fas_hangup (fip)
register struct fas_internals	*fip REG_SI;
{
	int	old_level;

	old_level = SPLINT ();

	if (fip->device_flags & DF_DO_HANGUP)
	{
		/* do the hangup */
		MCR &= ~(fip->modem.m.ei | fip->modem.m.eo
				| fip->flow.m.ic | fip->flow.m.hc);
		FAS_FIRST_OUTB (fip, MDM_CTL_PORT, MCR |= fip->modem.m.di);
		fip->device_flags &= ~DF_DO_HANGUP;
		(void) timeout (fas_hangup, (caddr_t) fip,
						(HANGUP_TIME) * (HZ) / 1000);
	}
	else
	{
		/* If there was a waiting dialin process on this
		   port, reopen the physical device.
		*/
		if (fip->o_state & OS_WAIT_OPEN)
		{
			/* allow pending tty interrupts */
			(void) SPLWRK ();
			(void) SPLINT ();

			fas_open_device (fip);

			/* allow pending tty interrupts */
			(void) SPLWRK ();
			(void) SPLINT ();
		}
		RELEASE_DEVICE_LOCK (fip);
	}

	(void) splx (old_level);
}

/* Main timeout function.
   (called via timeout () at an OS dependent SPL)
*/
STATIC void
fas_timeout (fip)
register struct fas_internals	*fip REG_SI;
{
	int	old_level;

	old_level = SPLINT ();

	/* handle break request */
	if (fip->device_flags & DF_DO_BREAK)
	{
		/* set up break request flags */
		FAS_FIRST_OUTB (fip, LINE_CTL_PORT,
					LCR |= LC_SET_BREAK_LEVEL);
		fip->device_flags &= ~(DF_DO_BREAK | DF_GUARD_TIMEOUT);
		(void) timeout (fas_timeout, (caddr_t) fip,
						(BREAK_TIME) * (HZ) / 1000);
		(void) splx (old_level);
		return;
	}

	/* reset break state */
	if ((fip->device_flags & DF_XMIT_BREAK)
		&& (LCR & LC_SET_BREAK_LEVEL))
	{
		FAS_FIRST_OUTB (fip, LINE_CTL_PORT, LCR &= ~LC_SET_BREAK_LEVEL);
		fip->device_flags |= DF_GUARD_TIMEOUT;
		fip->timeout_idx = timeout (fas_timeout, (caddr_t) fip,
						fas_speed_ptr [BT_SELECT]
							[fip->cflag & CBAUD]
							.i.ctime);
		(void) splx (old_level);
		return;
	}

	fip->device_flags &= ~(DF_GUARD_TIMEOUT | DF_XMIT_BREAK);

	/* restart output */
	fas_xproc (fip);

	if ((fip->flow_flags & FF_OUTPUT_BUSY)
			&& !fip->xmit_ring_cnt
			&& !(fip->device_flags & DF_XMIT_BUSY))
	{
		fip->flow_flags &= ~FF_OUTPUT_BUSY;
		fip->tty->t_state &= ~BUSY;
		/* check half duplex output flow control */
		if ((fip->flow_flags & (FF_HDX_HANDSHAKE | FF_HDX_STARTED))
				== (FF_HDX_HANDSHAKE | FF_HDX_STARTED))
		{
			FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR &= ~fip->flow.m.hc);
			fip->flow_flags &= ~FF_HDX_STARTED;
		}
		EVENT_SCHED (fip, EF_DO_XXFER);
	}

	(void) wakeup ((caddr_t) &(fip)->device_flags);
	(void) splx (old_level);
}

/* Re-enable receiver data interrupts.
   (called at SPLINT)
*/
STATIC void
fas_rdi_enable (fip)
register struct fas_internals	*fip REG_SI;
{
	/* flush receiver register */
	if (DEVICE_TYPE == TYPE_NS16550A)
		FAS_FIRST_OUTB (fip, NS_FIFO_CTL_PORT, FCR | NS_FIFO_CLR_RECV);
	else if (DEVICE_TYPE == TYPE_I82510)
	{
		FAS_FIRST_OUTB (fip, I_BANK_PORT, I_BANK_1);
		FAS_OUTB (fip, I_RCM_PORT, I_FIFO_CLR_RECV);
		FAS_OUTB (fip, I_BANK_PORT, I_BANK_0);
	}
	(void) FAS_FIRST_INB (fip, RCV_DATA_PORT);
	if (FAS_INB (fip, LINE_STATUS_PORT) & LS_RCV_AVAIL)
		(void) FAS_INB (fip, RCV_DATA_PORT);

	FAS_OUTB (fip, INT_ENABLE_PORT, IER |= IE_RECV_DATA_AVAILABLE);
	fip->device_flags |= DF_RDI_ENABLED;
}

/* Start software input flow control.
   (called at SPLINT)
*/
STATIC void
fas_send_xon (fip)
register struct fas_internals	*fip REG_SI;
{
	fip->flow_flags &= ~FF_SWI_STOPPED;
	fip->flow_flags ^= FF_SW_FC_REQ;
	if (fip->flow_flags & FF_SW_FC_REQ)
	{
	    if (fip->flow_flags & FF_CARRIER_ON)
	    {
		/* start input */
		fip->flow_flags |= FF_OUTPUT_BUSY;
		fip->tty->t_state |= TTXON | BUSY;
		/* check half duplex output flow control */
		if ((fip->flow_flags & (FF_HDX_HANDSHAKE | FF_HDX_STARTED))
			== FF_HDX_HANDSHAKE)
		{
			FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR |= fip->flow.m.hc);
			fip->flow_flags |= FF_HDX_STARTED;
		}
		fas_xproc (fip);
	    }
	    else
		fip->flow_flags &= ~FF_SW_FC_REQ;
	}
	else
	{
		/* there is a yet unprocessed request for XOFF, so
		   we just cancel it and don't need to send an XON
		*/
		fip->tty->t_state &= ~TTXOFF;
		if ((fip->flow_flags & FF_OUTPUT_BUSY)
				&& !fip->xmit_ring_cnt
				&& !(fip->device_flags
						& (DF_XMIT_BUSY
							| DF_GUARD_TIMEOUT
							| DF_XMIT_BREAK)))
		{
			fip->flow_flags &= ~FF_OUTPUT_BUSY;
			fip->tty->t_state &= ~BUSY;
			/* check half duplex output flow control */
			if ((fip->flow_flags & (FF_HDX_HANDSHAKE
							| FF_HDX_STARTED))
				== (FF_HDX_HANDSHAKE | FF_HDX_STARTED))
			{
				FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR &= ~fip->flow.m.hc);
				fip->flow_flags &= ~FF_HDX_STARTED;
			}
			EVENT_SCHED (fip, EF_DO_XXFER);
		}
	}
}

/* Stop software input flow control.
   (called at SPLINT)
*/
STATIC void
fas_send_xoff (fip)
register struct fas_internals	*fip REG_SI;
{
	fip->flow_flags |= FF_SWI_STOPPED;
	fip->flow_flags ^= FF_SW_FC_REQ;
	if (fip->flow_flags & FF_SW_FC_REQ)
	{
	    if (fip->flow_flags & FF_CARRIER_ON)
	    {
		/* stop input */
		fip->flow_flags |= FF_OUTPUT_BUSY;
		fip->tty->t_state |= TTXOFF | BUSY;
		/* check half duplex output flow control */
		if ((fip->flow_flags & (FF_HDX_HANDSHAKE | FF_HDX_STARTED))
			== FF_HDX_HANDSHAKE)
		{
			FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR |= fip->flow.m.hc);
			fip->flow_flags |= FF_HDX_STARTED;
		}
		fas_xproc (fip);
	    }
	    else
		fip->flow_flags &= ~FF_SW_FC_REQ;
	}
	else
	{
		/* there is a yet unprocessed request for XON, so
		   we just cancel it and don't need to send an XOFF
		*/
		fip->tty->t_state &= ~TTXON;
		if ((fip->flow_flags & FF_OUTPUT_BUSY)
				&& !fip->xmit_ring_cnt
				&& !(fip->device_flags
						& (DF_XMIT_BUSY
							| DF_GUARD_TIMEOUT
							| DF_XMIT_BREAK)))
		{
			fip->flow_flags &= ~FF_OUTPUT_BUSY;
			fip->tty->t_state &= ~BUSY;
			/* check half duplex output flow control */
			if ((fip->flow_flags & (FF_HDX_HANDSHAKE
							| FF_HDX_STARTED))
				== (FF_HDX_HANDSHAKE | FF_HDX_STARTED))
			{
				FAS_FIRST_OUTB (fip, MDM_CTL_PORT,
						MCR &= ~fip->flow.m.hc);
				fip->flow_flags &= ~FF_HDX_STARTED;
			}
			EVENT_SCHED (fip, EF_DO_XXFER);
		}
	}
}

/* Compute the port access control value. */
STATIC uint
fas_make_ctl_val (fip, unit, num)
register struct fas_internals	*fip REG_SI;
uint	unit;
uint	num;
{
	register uint	val REG_DI;
	register uint	mask REG_BX;
	uint	i;

	if (fip->device_flags & DF_CTL_FIRST)
		return (fas_ctl_val [unit]);

	if (fip->device_flags & DF_CTL_EVERY)
	{
		for (i = 8, mask = fas_ctl_val [unit],
				val = fas_ctl_val [unit] << 8; i; --i)
		{
			if (mask & 0x100)
			{
				if (num & 0x01)
					val ^= 0x100;
				num >>= 1;
			}
			mask >>= 1;
			val >>= 1;
		}
		return (val);
	}
	return (0);
}

/* Initialize memory with zeros. */
STATIC void
fas_mem_zero (ptr, size)
register unchar	*ptr REG_SI;
uint	size;
{
	register unchar	*limptr REG_DI;

	for (limptr = ptr + size; ptr < limptr; ++ptr)
		*ptr = 0;
}

/* Test device thoroughly. */
STATIC uint
fas_test_device (fip)
register struct fas_internals	*fip REG_SI;
{
	register unchar	*cptr REG_DI;
	uint	errcode;
	register uint	delay_count REG_BX;
	uint	rep_count, i;
	n_unchar	lsr, msr;
	static uint	lcrval [3] =
	{
		LC_WORDLEN_8,
		LC_WORDLEN_8 | LC_ENABLE_PARITY,
		LC_WORDLEN_8 | LC_ENABLE_PARITY | LC_EVEN_PARITY
	};

	/* make sure FIFO is off */
	FAS_FIRST_OUTB (fip, NS_FIFO_CTL_PORT, NS_FIFO_CLEAR_CMD);
	FAS_OUTB (fip, I_BANK_PORT, I_BANK_2);
	FAS_OUTB (fip, I_IDM_PORT, I_FIFO_CLEAR_CMD);
	FAS_OUTB (fip, I_BANK_PORT, I_BANK_0);

	/* set counter divisor */
	FAS_OUTB (fip, LINE_CTL_PORT, LC_ENABLE_DIVISOR);
	FAS_OUTB (fip, DIVISOR_LSB_PORT, fas_speed_ptr [BT_SELECT]
							[B0].div.b.low);
	FAS_OUTB (fip, DIVISOR_MSB_PORT, fas_speed_ptr [BT_SELECT]
							[B0].div.b.high);
	FAS_OUTB (fip, LINE_CTL_PORT, 0);

	/* switch to local loopback */
	FAS_OUTB (fip, MDM_CTL_PORT, MC_SET_LOOPBACK);

	errcode = 0;

	/* wait until the transmitter register is empty */
	FAS_EVERY_CTL (fip, LINE_STATUS_PORT);
	for (delay_count = TEST_DELAY_LOOPS;
		delay_count && (~FAS_SAME_INB (fip, LINE_STATUS_PORT)
				& (LS_XMIT_AVAIL | LS_XMIT_COMPLETE));
		--delay_count)
		;

	if (!delay_count)
		errcode = 1;

	if (!errcode)
	{
	    /* clear flags */
	    (void) FAS_INB (fip, RCV_DATA_PORT);
	    (void) FAS_SAME_INB (fip, RCV_DATA_PORT);
	    (void) FAS_INB (fip, LINE_STATUS_PORT);

	    /* make sure there are no more characters in the
	       receiver register
	    */
	    FAS_EVERY_CTL (fip, LINE_STATUS_PORT);
	    for (delay_count = TEST_DELAY_LOOPS;
		delay_count && !(FAS_SAME_INB (fip, LINE_STATUS_PORT)
						& LS_RCV_AVAIL);
		--delay_count)
		;

	    if (delay_count)
		(void) FAS_INB (fip, RCV_DATA_PORT);

	    for (rep_count = TEST_CHAR_LOOPS; rep_count; --rep_count)
	    {
		for (i = 0; i < 3; ++i)
		{
			/* test pattern */
			cptr = (unchar *) "\377\125\252\045\244\0\0";

			/* check LSR flags */
			if (FAS_INB (fip, LINE_STATUS_PORT)
					!= (LS_XMIT_AVAIL | LS_XMIT_COMPLETE))
			{
				errcode = 2;
				break;
			}

			/* test transmitter and receiver
			   with different line settings
			*/
			FAS_OUTB (fip, LINE_CTL_PORT, lcrval [i]);

			/* send first test pattern */
			FAS_OUTB (fip, XMT_DATA_PORT, *cptr);

			/* wait until the transmitter holding register
			   is empty
			*/
			FAS_EVERY_CTL (fip, LINE_STATUS_PORT);
			for (delay_count = TEST_DELAY_LOOPS;
					delay_count
					  && !((lsr = FAS_SAME_INB (fip,
						    LINE_STATUS_PORT))
							& LS_XMIT_AVAIL);
				--delay_count)
				;

			if (!delay_count)
			{
				errcode = 3;
				break;
			}

			/* check LSR flags */
			if (lsr != LS_XMIT_AVAIL)
			{
				errcode = 2;
				break;
			}

			do
			{
				if (*cptr)
				{
					/* send next test pattern */
					FAS_OUTB (fip, XMT_DATA_PORT, *(cptr + 1));

					/* check LSR flags */
					if (FAS_INB (fip, LINE_STATUS_PORT))
					{
						errcode = 2;
						break;
					}
				}

				/* wait until the test pattern is received */
				FAS_EVERY_CTL (fip, LINE_STATUS_PORT);
				for (delay_count = TEST_DELAY_LOOPS;
					delay_count
					  && !((lsr = FAS_SAME_INB (fip,
							LINE_STATUS_PORT))
								& LS_RCV_AVAIL);
					--delay_count)
					;

				if (!delay_count)
				{
					errcode = 4;
					break;
				}

				/* check LSR flags */
				if ((lsr & LS_RCV_INT) != LS_RCV_AVAIL)
				{
					errcode = 5;
					break;
				}

				/* check test pattern */
				if (FAS_INB (fip, RCV_DATA_PORT) != *cptr)
				{
					errcode = 6;
					break;
				}

				/* check LSR flags */
				if (FAS_INB (fip, LINE_STATUS_PORT)
							& LS_RCV_INT)
				{
					errcode = 5;
					break;
				}

				/* wait until the transmitter register
				   is empty
				*/
				if (*cptr)
				{
				    FAS_EVERY_CTL (fip, LINE_STATUS_PORT);
				    for (delay_count = TEST_DELAY_LOOPS;
					delay_count
					  && !((lsr = FAS_SAME_INB (fip,
						    LINE_STATUS_PORT))
							& LS_XMIT_AVAIL);
					--delay_count)
					;

				    if (!delay_count)
				    {
					errcode = 7;
					break;
				    }

				    /* check LSR flags */
				    if (lsr != LS_XMIT_AVAIL)
				    {
					errcode = 8;
					break;
				    }
				}
				else
				{
				    FAS_EVERY_CTL (fip, LINE_STATUS_PORT);
				    for (delay_count = TEST_DELAY_LOOPS;
					delay_count
					  && !((lsr = FAS_SAME_INB (fip,
						    LINE_STATUS_PORT))
							& LS_XMIT_COMPLETE);
					--delay_count)
					;

				    if (!delay_count)
				    {
					errcode = 7;
					break;
				    }

				    /* check LSR flags */
				    if (lsr != (LS_XMIT_AVAIL
							| LS_XMIT_COMPLETE))
				    {
					errcode = 8;
					break;
				    }
				}
			} while (*((ushort *) (cptr++)));

			if (errcode)
				break;
		}

		if (errcode)
			break;
	    }
	}

	if (!errcode)
	{
	    /* clear delta bits */
	    (void) FAS_INB (fip, MDM_STATUS_PORT);

	    for (rep_count = TEST_CTL_LOOPS; rep_count; --rep_count)
	    {
		/* test pattern */
		cptr = (unchar *) "\005\142\012\237\006\130\011\257\017\361\0\017\0\0";

		for (; *((ushort *) cptr); cptr += 2)
		{
			/* test modem control and status lines */
			FAS_OUTB (fip, MDM_CTL_PORT, *cptr | MC_SET_LOOPBACK);

			/* wait for delta flags */
			FAS_EVERY_CTL (fip, MDM_STATUS_PORT);
			for (delay_count = TEST_DELAY_LOOPS;
				delay_count
					&& !((msr = FAS_SAME_INB (fip,
							    MDM_STATUS_PORT))
							& MS_ANY_DELTA);
				--delay_count)
				;

			if (!delay_count)
			{
				errcode = 9;
				break;
			}

			/* check MSR flags */
			if (msr != *(cptr + 1))
			{
				errcode = 9;
				break;
			}

			/* check whether delta flags are cleared */
			if (FAS_SAME_INB (fip, MDM_STATUS_PORT)
					!= (msr & ~MS_ANY_DELTA))
			{
				errcode = 9;
				break;
			}
		}

		if (errcode)
			break;
	    }
	}

	/* switch back to normal operation */
	FAS_OUTB (fip, MDM_CTL_PORT, 0);

	return (errcode);
}

#if defined (NEED_PUT_GETCHAR)
int
FASPUTCHAR (arg)
unchar	arg;
{
	register struct	fas_internals	*fip REG_SI;

	if (!fas_is_initted)
		fasinit ();

	fip = &fas_internals [0];
	if (fip->device_flags & DF_DEVICE_CONFIGURED)
	{
		if (arg == '\n')
			(void) FASPUTCHAR ('\r');
		FAS_CTL (fip, LINE_STATUS_PORT);
		while (!(FAS_SAME_INB (fip, LINE_STATUS_PORT) & LS_XMIT_AVAIL))
			;
		FAS_OUTB (fip, XMT_DATA_PORT, arg);
	}
	return (0);
}

int
FASGETCHAR ()
{
	register struct	fas_internals	*fip REG_SI;

	if (!fas_is_initted)
		fasinit ();

	fip = &fas_internals [0];
	if ((fip->device_flags & DF_DEVICE_CONFIGURED)
		&& (FAS_FIRST_INB (fip, LINE_STATUS_PORT) & LS_RCV_AVAIL))
		return (FAS_INB (fip, RCV_DATA_PORT));
	else
		return (-1);
}
#endif

#if defined (NEED_INIT8250)
/* Reset the requested port to be used directly by a DOS process.
   (DosMerge only)
*/
int
init8250 (port, ier)
register ushort	port REG_BX;
ushort	ier;	/* ier not used in this stub */
{
	register struct fas_internals	*fip REG_SI;
	register uint	unit REG_DI;
	int	old_level;

	/* See if the port address matches a port that is used by
	   the FAS driver.
	*/
	for (unit = 0; unit < fas_physical_units; ++unit)
		if (port == fas_port [unit])
			break;

	if (unit >= fas_physical_units)
		return (-1);	/* port didn't match */

	fip = fas_internals_ptr [unit];

	old_level = SPLINT ();

	FAS_FIRST_OUTB (fip, INT_ENABLE_PORT, IER = IE_NONE);
	fip->device_flags &= ~(DF_RDI_ENABLED | DF_MSI_ENABLED | DF_MSI_NOISE);

	FAS_OUTB (fip, MDM_CTL_PORT, MCR &= ~(fip->flow.m.ic | fip->flow.m.hc));

	if (DEVICE_TYPE == TYPE_NS16550A)
		FAS_OUTB (fip, NS_FIFO_CTL_PORT, NS_FIFO_CLEAR_CMD);
	else if (DEVICE_TYPE == TYPE_I82510)
	{
		FAS_OUTB (fip, I_BANK_PORT, I_BANK_1);
		FAS_OUTB (fip, I_TCM_PORT, I_FIFO_CLR_XMIT);
		FAS_OUTB (fip, I_RCM_PORT, I_FIFO_CLR_RECV);
		FAS_OUTB (fip, I_BANK_PORT, I_BANK_2);
		FAS_OUTB (fip, I_IDM_PORT, I_FIFO_CLEAR_CMD);
		FAS_OUTB (fip, I_BANK_PORT, I_BANK_0);
	}

	(void) FAS_INB (fip, MDM_STATUS_PORT);
	(void) FAS_INB (fip, RCV_DATA_PORT);
	if (FAS_INB (fip, LINE_STATUS_PORT) & LS_RCV_AVAIL)
		(void) FAS_INB (fip, RCV_DATA_PORT);
	(void) FAS_INB (fip, INT_ID_PORT);
	(void) splx (old_level);
	return (0);
}
#endif
